//-------+---------+---------+---------+---------+---------+---------+--------=
//
// File: lastz.c
//
//----------

static char* programName            = "lastz";
static char* programVersionMajor    = VERSION_MAJOR;
static char* programVersionMinor    = VERSION_MINOR;
static char* programVersionSubMinor = VERSION_SUBMINOR;
static char* programRevisionDate    = REVISION_DATE;

static char* svnRevisionNumber      = SUBVERSION_REV;

//----------
//
// lastz-- Local Alignment Search Tool, blastZ-like
//	Find pairwise local alignments between a target DNA sequence and a series
//	of query sequences.  Query sequences can be DNA or quantum DNA.
//
// GOAL 1: Align sequences as large as 250M bases (about the size of human
//         chromsome 1) on a workstation with 32-bit addressing and 2G byte of
//         memory, with a seed up to weight 24 (equivalent to a 12-mer).
//
// GOAL 2: It should be relatively easy to try out different alignment
//         strategies.
//
//----------
//
// Algorithmic Overview
//
// Caveat:  many details are left out or glosssed over in this description.  In
// some cases accuracy has been sacrificed for clarity.
//
// The algorithm consists of the following stages:
//	1 (SEED)   discovery of short near-matches ("seed hits")
//	2 (HSP)    extension to ungapped High-scoring Segment Pairs (HSPs)
//	3 (CHAIN)  reduction of HSPs to the best-scoring "chain"
//	4 (ANCHOR) reduction of (remaining) HSPs to single positions ("anchors")
//	5 (ALIGN)  extension of anchors to gapped alignments
//	6 (INTERP) "interpolation" between alignments at a higher sensitivity
//
// Most stages are optional.  For example, you can leave out chaining, or you
// can stop as soon as HSPs are found.  However, there are some dependencies
// between stages.  SEED and HSP are performed together and are required for
// anything else, ANCHOR and ALIGN are performed together, and INTERP requires
// ALIGN.  INTERP can be thought of as repeating the first five stages on
// whatever failed to align.
//
// The process is repeated, (mostly) independently, for every query sequence.
// The exceptions are a seed position table is built once, and dynamic masking
// has interactions between queries. (These are described below).
//
// The SEED stage makes use of a large table containing the positions (in
// the target sequence) of every possible seed bit pattern.  For this discussion
// seed can be thought of as 12-mers, but we support more general seeds (see
// seeds.c).  The table is indexed by the seed bit pattern, which for an 12-mer
// is a 24 bit value.  For each seed value the table gives a list of all
// positions where that seed value can be found.  The table requires 4*(L+4^W)
// bytes, where W is the seed length and L is the sequence length.  For W=12
// and L = 250MB, this is a little larger than 1GB.  For details on how the
// table is stored, see pos_table.c.
//
// Having created the position table, the SEED stage scans a query sequence and
// looks up matches ("seed hits") for every query seed.  As each seed hit is
// found, the HSP stage is performed, extending it along the diagonal in both
// directions until the score drops off.  HSPs that do not meet the score
// threshold are discarded (strictly speaking, they are not HSPs).  To improve
// performance, we keep track of how far we have progressed along each diagonal,
// and quickly discard any seed hits that fall into previously identified HSPs.
// Since a complete diagonal tracking array could be huge (for two 250M base
// sequences it would be 4 * 500M = 2G), we use a much smaller array and hash
// diagonals into its index space.  This results in a small loss of sensitivity,
// as hash collisions between diagonals can cause seed hits on other diagonals
// to be hidden by HSPs on other diagonals.  See seed_search.c and diag_hash.c.
//
// The CHAIN stage finds the highest scoring series of HSPs in which each HSP
// begins strictly before the start of the next.  All HSPs not on this chain
// are discarded.  This is useful when the query and target are known to be
// syntenic.  The algorithm processes the HSPs in order along the target, build
// chains by adding the next HSP to the best previous viable chain.  See
// chain.c.
//
// The ANCHOR stage reduces each HSP to a single point.  A constant-width
// window is slid across the HSP and the midpoint of the highest-scoring window
// is chosen as the anchor.
//
// The ALIGN stage extends anchors into gapped alignments.  One-sided extension
// is performed in two directions from the anchor point, the two resulting
// alignments are joined at the anchor, and if the score is high enough, this
// becomes an alignment in the output file.  Extension is computed per the
// standard (but optimized) 3-state affine gap dynamic programming recurrence.
// The DP matrix is evaluated only in a small region straddling the high-scoring
// path, with only enough memory to store the widest row of that region.  See
// gapped_extend.c.
//
// Dynamic masking is (optionally) performed during the ALIGN stage.  We keep a
// count of how many times each target base has been in an alignment.  When a
// base reaches a threshold, it is assumed to be a repeat and is masked from the
// sequence (by chaing it to an 'x').  This only affects subsequent queries in
// the same run.  See masking.c.
//
// The INTERP stage repeats all of the previous steps in the regions between
// those gapped alignments.  As of this writing the interpolation seed is a
// 7-mer exact match, for compatibility with BLASTZ.
//
// Input formats can be FASTA, nib, or qdna files, but other formats can be
// added without too much pain.  See sequences.c.
//
// A variety of pairwise output formats are supported.  The primary format is
// LAV for compatibility with BLASTZ.  See lav.c, gfa.c, axt.c and text_align.c.
// It is relatively easy to add other formats (although it would be even easier
// if this were written in an object-oriented language).
//
//----------

//----------
//
// other files
//
//----------

#include <stdlib.h>				// standard C stuff
#define  true  1
#define  false 0
#include <stdio.h>				// standard C i/o stuff
#include <string.h>				// standard C string stuff
#include <stdarg.h>				// standard C variable argument list stuff
#include <math.h>				// standard C math stuff
#include <time.h>				// standard C time stuff
#include "utilities.h"			// utility stuff
#include "dna_utilities.h"		// dna/scoring stuff
#include "sequences.h"			// sequence stuff
#include "seeds.h"				// seed strategy stuff
#include "pos_table.h"			// position table stuff
#include "seed_search.h"		// seed hit search stuff
#include "quantum.h"			// quantum DNA search stuff
#include "segment.h"			// segment table management stuff
#include "chain.h"				// segment chaining stuff
#include "gapped_extend.h"		// gapped alignment stuff
#include "tweener.h"			// interpolated alignment stuff
#include "masking.h"			// dynamic masking stuff
#include "infer_scores.h"		// scoring inference stuff
#include "edit_script.h"		// alignment edit script stuff
#include "diag_hash.h"			// diagonals hashing stuff
#include "identity_dist.h"		// identity distribution stuff
#include "coverage_dist.h"		// query coverage distribution stuff
#include "gfa.h"				// gfa alignment format stuff
#include "lav.h"				// lav alignment format stuff
#include "axt.h"				// axt alignment format stuff
#include "maf.h"				// maf alignment format stuff
#include "text_align.h"			// textual alignment format stuff

#define  lastz_owner			// (make this owner of program-wide globals)
#include "lastz.h"				// lastz program-wide stuff

#define helpout stdout			// stream to write help messages to

//----------
//
// stuff for crude profiling
//
//----------

//#define useStandardClock

#ifndef dbgTiming
#define dbg_timing_sub(v) ;
#define dbg_timing_add(v) ;
#define dbg_timing_copy(dst,src) ;
#define dbg_timing_report(v,s) ;
#endif // not dbgTiming

#ifdef dbgTiming

#ifdef useStandardClock
#define read_clock() clock()
#define clocksPerSec CLOCKS_PER_SEC
#endif // useStandardClock

#ifndef useStandardClock
#include <sys/time.h>
#define read_clock() microsec_clock()
#define clocksPerSec 1000000
static u64 microsec_clock (void);
static u64 microsec_clock (void)
	{
	static int		failed = false;
	struct timeval	time1;
	int				err;

	if (failed)	return 0;	// (previous call to gettimeofday has failed)
	err = gettimeofday (&time1, NULL);
	if (err != 0) { failed = true;  return 0; }

	return (((u64) time1.tv_sec) * 1000000) + time1.tv_usec;
	}
#endif // not useStandardClock

s64 debugClockTotal = 0,
    debugClockSeq1 = 0,
    debugClockSeq2 = 0,
    debugClockPosTable = 0,
    debugClockQueryTotal = 0,
    debugClockSegTable = 0,
    debugClockChaining = 0,
    debugClockGappedExtend = 0,
    debugClockInterpolation = 0,
    debugClockOutput = 0;

#define dbg_timing_sub(v)  { v -= (s64) read_clock();  }
#define dbg_timing_add(v)  { v += (s64) read_clock();  }
#define dbg_timing_copy(dst,src) { dst = src; }

#define dbg_timing_report(v,s) { fprintf(stderr,"%-26s %.3f\n",s":",((float)(v))/clocksPerSec); }
#endif // dbgTiming

//----------
//
// private global data
//
//----------

static int showProgress = false;

// we keep two copies of the control data, one for the primary alignment, and
// another for alignments used to infer a scoring set;  for historical reasons
// we call these "lz" (short for lastz) and "iz" (short for inferz);  we keep a
// pointer to whichever of these is 'active' at any given time

static control  lzParams;
static control  izParams;
static control* currParams;

// command line options

// defaults

static const control defaultParams =
	{
	NULL,NULL,							// seq1, seq1Filename
	NULL,NULL,							// seq2, seq2Filename

	false,false,						// inferScores, inferOnly
	{
	NULL,								// ic.inferFileName
	100,true,							// ic.inferScale, ic.writeAsInt
	ratioNone,							// ic.hspThresholdIsRatio,
	ratioNone,							// ic.gappedThresholdIsRatio
	ratioNone,							// ic.gapOpenIsRatio
	ratioNone,							// ic.gapExtendIsRatio
	30,0,								// ic.subIterations, ic.gapIterations
	false,								// ic.idIsPercentile
	},
	false,								// selfCompare
	true,								// doSeedSearch
	(s8*) upper_nuc_to_bits,			// charToBits
	1,									// whichStrand
	false,								// backward
	1,									// step

	NULL,								// hitSeed (default is defaultSeedString)
	28,									// maxIndexBits
	1,									// withTrans
	false,								// noHitFiltering
	0,0,								// twinMinSpan, twinMaxSpan (trumped
										// .. by defaultTwinsYes, etc. below)
	hitSimple,							// basicHitType
	-1,0,								// minMatches, maxTransversions

	gfexXDrop,							// gfExtend
	false,								// mergeAnchors
	false,0,0,							// chain, chainDiag, chainAnti
	true,								// gappedExtend

	NULL,NULL,							// scoring, maskedScoring
	0,									// xDrop
	0,									// yDrop
	{'S',3000,0,0},						// hspThreshold
	{'S',0,   0,0},						// gappedThreshold
	true,false,							// entropicHsp, reportEntropy
	-1,									// mirror (not used, default hardcoded)
	80*1024*1024,						// tracebackMem
	0,									// dynamicMasking
	false,								// reportCensus
	0.0,1.0,							// minIdentity, maxIdentity
	0.0,1.0,							// minCoverage, maxCoverage
	fmtLav,false,						// outputFormat, outputNeedsTrueLen

	0,									// innerThreshold
	NULL,								// innerSeed
	20000,								// innerWindow

	false,								// queryIsQuantum
	-1,									// ballScore
	NULL,								// coding

	true,								// lajCompatible
	0,									// textContext
	NULL,								// args
	0,									// verbosity
	false,								// reportTiming
	false,								// reportStats
	false,								// showStats
	NULL,								// statsFile
	NULL,								// statsFileName
	0									// showPosTable
	};

static const char* defaultSeedString = seed_12of19;

static const int   defaultTwinsYes   = false;
static const int   defaultTwinMinGap = 0;
static const int   defaultTwinMaxGap = 10;

static int dbgShowParams   = false;
static int dbgShowHsps     = false;
static int dbgShowAnchors  = false;
static int dbgInhibitSegmentReduction = false;
static int dbgQueryProgress = 0;

#define innerWordSize	7				// word size for inner alignment seed
										// .. hits

static const float defaultBallScoreFactor = 0.75;

// anchors--
//	Whenever the process will go beyond just finding gap-free extensions, the
//	segments that will become anchors (e.g. HSPs) are collected in this table
//	instead of being written to the console.

#define numDefaultAnchors 4000

static segtable* anchors      = NULL;
static segtable* savedAnchors = NULL;

// miscellany

#define chainScale 100

//----------
//
// prototypes for private functions
//
//----------

int main (int argc, char** argv);

static score chain_connect_penalty   (segment* seg1, segment* seg2, int scale);
static void  remove_interval_seeds   (unspos b, unspos e, void* info);
static u32   report_hsps             (void* info,
                                      unspos pos1, unspos pos2, unspos length,
                                      score s);
static u32   collect_hsps            (void* info,
                                      unspos pos1, unspos pos2, unspos length,
                                      score s);
static void  parse_options           (int argc, char** argv,
                                      control* lzParams, control* izParams);
static void  print_params            (FILE* f);

static void  read_control_file_by_name (char* name, control* params);
static void  read_control_file       (FILE* f, char* name, control* params);

static void  print_quantum_coding    (qcode* coding);
static void  print_job_footer        (void);
static void  print_header            (void);
static void  print_align_list        (alignel* alignList);
static void  print_match             (unspos pos1, unspos pos2, unspos length,
                                      score s);
static void  print_m_stanza          (census* cen);
static void  print_census_stanza     (census* cen);
static void  print_x_stanza          (unspos numMasked);
static void  print_generic           (FILE* f, const char* format, ...);
static void  print_options           (void);

static void  print_match_composition (FILE* f,
                                      seq* seq1, unspos pos1,
                                      seq* seq2, unspos pos2,
                                      unspos length,
                                      score s,
                                      seed* hitSeed, u32 step);
static void  dump_match              (FILE* f,
                                      seq* seq1, unspos pos1,
                                      seq* seq2, unspos pos2,
                                      unspos length);

static void  lastz_zero_stats        (void);
static void  lastz_show_stats_before (FILE* f);
static void  lastz_show_stats        (FILE* f);

//----------
//
// lastz--
//  Main program
//
//----------

int main
   (int				argc,
	char**			argv)
	{
	FILE*			statsF        = NULL;
	seq*			target        = NULL;
	seq*			query         = NULL;
	postable*		targPositions = NULL;
	u8*				targetRev     = NULL;
	tback*			traceback     = NULL;
	census*			targCensus    = NULL;
	hitprocessor	hitProc;
	void*			hitProcInfo;
	time_t			startClock, endClock;
	unspos			coverageLimit;
	int				hspsAreAdaptive;
	int				queryExists;
	int				numQueries;

	dbg_timing_sub (debugClockTotal);

	debug = 0;

	lastz_zero_stats ();
	pos_table_zero_stats ();
	seed_search_zero_stats ();
	quantum_zero_stats ();
	chain_zero_stats ();
	gapped_extend_zero_stats ();
	tweener_zero_stats ();
	masking_zero_stats ();
	infer_scores_zero_stats ();

	//////////
	// fetch arguments
	//////////

	currParams = NULL;
	parse_options (argc, argv, &lzParams, &izParams);
	currParams = &lzParams;

	// open stats file

	if (currParams->statsFileName != NULL)
		{
		currParams->statsFile = fopen_or_die (currParams->statsFileName, "wt");
		free_if_valid ("stats file name", currParams->statsFileName);
		}

	if (currParams->showStats)
		statsF = (currParams->statsFile != NULL)? currParams->statsFile : stderr;

	if ((currParams->inferScores) && (statsF != NULL))
		izParams.statsFile = statsF;

	//////////
	// open the input files
	//
	// We also load the first sequence from the input files here.  This allows
	// us to check and report problems (such as an extra target or an empty
	// query) immediately, rather than wasting what might turn out to be a lot
	// of processing time.
	//////////

	startClock = clock();

	// open and load target

	memory_checkpoint ("[[* Opening Files ]]\n");

	dbg_timing_sub (debugClockSeq1);

	if (currParams->verbosity >= 5)
		fprintf (stderr, "opening \"%s\"\n", currParams->seq1Filename);

	target = open_sequence_file (currParams->seq1Filename, seq_type_unknown,
	                             currParams->outputNeedsTrueLen);
	currParams->seq1 = target;

	if (!load_sequence (target))
		suicidef ("%s contains no sequence", target->fileName);
	if (target->fileType == seq_type_qdna)
		suicidef ("quantum DNA not supported for target (%s)", target->fileName);
	if ((currParams->maskedScoring != NULL) && (!currParams->maskedScoring->rowsAreDna))
		suicidef ("row scores are not allowed to be for quantum DNA");
	if (another_sequence (target))
		suicidef ("%s contains more than one sequence", target->fileName);

	if ((currParams->gappedExtend)
	 || ((currParams->inferScores) && (izParams.gappedExtend)))
		targetRev = (u8*) reverse_of_string ((char*) target->v, target->len);

	dbg_timing_add (debugClockSeq1);

	if ((currParams->dynamicMasking > 0) || (currParams->reportCensus))
		targCensus = new_census (target->len, currParams->dynamicMasking);

	// now that we know the target length, set the hsp and gapped thresholds
	// if they are a percentage of the target length

	resolve_score_thresh (&currParams->hspThreshold,    target->len);
	resolve_score_thresh (&currParams->gappedThreshold, target->len);

	if (currParams->inferScores)
		{
		resolve_score_thresh (&izParams.hspThreshold,    target->len);
		resolve_score_thresh (&izParams.gappedThreshold, target->len);
		}

	// open and load query

	if ((currParams->doSeedSearch) || (currParams->inferScores))
		{
		int fileType;
		if (currParams->verbosity >= 5)
			{
			if (currParams->seq2Filename != NULL)
				fprintf (stderr, "opening \"%s\"\n", currParams->seq2Filename);
			else
				fprintf (stderr, "opening unnamed query file\n");
			}

		dbg_timing_sub (debugClockSeq2);
		fileType = seq_type_unknown;
		if (currParams->queryIsQuantum)
			{
			fileType = seq_type_qdna;
			if (currParams->inferScores)
				suicide ("scoring inference cannot be performed with quantum DNA");
			}
		if (currParams->inferScores)
			query = open_rewindable_sequence_file (currParams->seq2Filename,
			                                       fileType,
			                                       currParams->outputNeedsTrueLen);
		else
			query = open_sequence_file            (currParams->seq2Filename,
			                                       fileType,
			                                       currParams->outputNeedsTrueLen);
		currParams->seq2 = query;

		if ((!currParams->queryIsQuantum) && (query->fileType == seq_type_qdna))
			suicidef ("%s does not contain quantum DNA", query->fileName);

		if (currParams->queryIsQuantum)
			{
			if ((currParams->minIdentity > 0) || (currParams->maxIdentity < 1))
				suicide ("identity filtering cannot be used with quantum DNA");
			if (currParams->outputFormat == fmtIdDist)
				suicide ("--format=identity cannot be used with quantum DNA");
			}

		if (!another_sequence (query))
			suicidef ("%s contains no sequence", query->fileName);

		dbg_timing_add (debugClockSeq2);
		}

	if (dbgShowParams)
		print_params (stderr);

	// check for bad combination of partitoned sequence vs other options

	if ((target->partition.p != NULL)
	 || (query->partition.p  != NULL))
		{
		if ((!currParams->inferOnly)
		 && ((currParams->outputFormat == fmtGfa)
		  || (currParams->outputFormat == fmtGfaNoScore)))
			suicide ("multi cannot be used with --gfa");

		if ((!currParams->inferOnly)
		 && ((currParams->outputFormat == fmtLav)
		  || (currParams->outputFormat == fmtLavComment)
		  || (currParams->outputFormat == fmtLavScore)
		  || (currParams->outputFormat == fmtLavText)))
			suicide ("multi cannot be used with --lav");
		}


	if ((target->partition.p != NULL) && (currParams->selfCompare))
		suicide ("multi cannot be used with --self");

	// allocate traceback memory

	if ((currParams->gappedExtend)
	 || ((currParams->inferScores) && (izParams.gappedExtend)))
		traceback = new_traceback (currParams->tracebackMem);

	//////////
	// build a position table for the target sequence
	//
	// If we are inferring scores, we'd like to share the same position table
	// for both inference and for the final alignment.  However, it is natural
	// to use a less sensitive seed or step during inference.  So in case the
	// seed or step is different, we use the inference control values here and
	// will rebuild the table after inference is finished, if needed.
	//////////

	dbg_timing_sub (debugClockPosTable);

	if (currParams->verbosity >= 1)
		fprintf (stderr, "building position table for %s\n", target->fileName);

	if (currParams->inferScores)
		targPositions = build_seed_position_table
						   (target, 0, target->len,
							currParams->charToBits, izParams.hitSeed,
							izParams.step);
	else
		targPositions = build_seed_position_table
						   (target, 0, target->len,
							currParams->charToBits, currParams->hitSeed,
							currParams->step);

	dbg_timing_add (debugClockPosTable);

	if (currParams->showPosTable != 0)
		{
		int countOnly = (currParams->showPosTable == 2);
		if (countOnly)
			printf ("seed-word counts table for %s:\n", currParams->seq1->fileName);
		else
			printf ("seed-word position table for %s:\n", currParams->seq1->fileName);
		dump_position_table (stdout, targPositions, currParams->hitSeed, countOnly);
		printf ("\n");
		}

	lastz_show_stats_before (statsF);
	pos_table_show_stats (statsF, targPositions);

	//////////
	// perform scoring inference, if requested
	//////////

	if (currParams->inferScores)
		{
		// switch control to the inference parameters

		currParams = &izParams;
		izParams.seq1Filename = lzParams.seq1Filename;
		izParams.seq1         = lzParams.seq1;
		izParams.seq2Filename = lzParams.seq2Filename;
		izParams.seq2         = lzParams.seq2;

		// perform the scoring inference (note that this will rewind the query
		// file)

		if (currParams->hspThreshold.t == 'S') coverageLimit = 0;
		else coverageLimit = currParams->hspThreshold.c;

		anchors = new_segment_table (numDefaultAnchors, coverageLimit);

		lzParams.scoring = drive_scoring_inference
		                       (currParams,
		                        target, targetRev, targPositions, query,
		                        traceback);

		// switch control back to the main parameters

		currParams = &lzParams;

		// fill in score-based parameters

		currParams->maskedScoring = masked_score_set (currParams->scoring);

		if (currParams->xDrop < 0)
			currParams->xDrop = 10 * currParams->scoring->sub['A'][currParams->scoring->colChars[0]];

		if (currParams->yDrop < 0)
			currParams->yDrop = currParams->scoring->gapOpen + 300 * currParams->scoring->gapExtend;

		// rebuild the target sequence position table if it would be different

		if ((currParams->doSeedSearch)
		 && (is_same_seed (izParams.hitSeed, currParams->hitSeed))
		 && (izParams.step != currParams->step))
			{
			free_position_table (targPositions);
			targPositions = build_seed_position_table
							   (target, 0, target->len,
								currParams->charToBits, currParams->hitSeed,
								currParams->step);
			}
		}

	//////////
	// perform the alignment of the query (or queries) to the target
	//////////

	dbg_timing_copy (debugClockQueryTotal, debugClockSeq2);
	dbg_timing_sub  (debugClockQueryTotal);
	numQueries = 0;

	// if the user doesn't want the alignment we can quit now

	if (((!currParams->inferScores) && (!currParams->doSeedSearch))
	 || (( currParams->inferScores) && (currParams->inferOnly)))
		goto clean_up;

	// initialize (or empty) the anchors list

	if (currParams->hspThreshold.t == 'S') coverageLimit = 0;
	else coverageLimit = currParams->hspThreshold.c;

	if (anchors == NULL)
		anchors = new_segment_table (numDefaultAnchors, coverageLimit);
	else
		{
		empty_segment_table (anchors);
		limit_segment_table (anchors, coverageLimit);
		}

	set_up_hit_processor (currParams, (targCensus!=NULL), &hitProc, &hitProcInfo);

	// search for hits in each query sequence

	hspsAreAdaptive = (currParams->hspThreshold.t != 'S');

	print_job_header ();
	if (currParams->queryIsQuantum)
		print_quantum_coding (currParams->coding);
	print_options ();

	while (true)
		{
		memory_checkpoint_1 ("[[* Query #%d (loading)]]\n", numQueries);

		dbg_timing_sub (debugClockSeq2);
		queryExists = load_sequence (query);
		dbg_timing_add (debugClockSeq2);
		if (!queryExists) break;
		numQueries++;

		memory_checkpoint_2 ("[[* Query #%d, %s (loaded) ]]\n",
		                     numQueries,
		                     (query->useFullNames)? query->header
		                                          : query->shortHeader);

		if (currParams->whichStrand < 0)
			revcomp_sequence (query);

		if ((dbgQueryProgress != 0)
		 && ((dbgQueryProgress == 1) || (numQueries % dbgQueryProgress == 1)))
		 	{
			fprintf (stderr, "processing query %d", numQueries);
			if (query->shortHeader != NULL)
				fprintf (stderr, ": %s", query->shortHeader);
			else if (query->header != NULL)
				fprintf (stderr, ": %s", query->header);
			fprintf (stderr, "\n");
			}

		// search for "forward" strand hits (and finish alignment for them if
		// hsps are non-adaptive)

		if (currParams->verbosity >= 1)
			fprintf (stderr, "searching for matches in %s%s\n",
							 query->fileName,
							 (currParams->whichStrand>=0)? ""
		                                             : ", (reverse strand)");
		start_one_strand (target, targPositions, query, true,
		                  hitProc, hitProcInfo);
		if (!hspsAreAdaptive)
			finish_one_strand (target, targetRev, targPositions, query, NULL,
							   traceback, targCensus);

		// search for reverse strand hits

		if (currParams->whichStrand > 0)
			{
			if (currParams->verbosity >= 1)
				fprintf (stderr, "searching for matches in %s, (reverse strand)\n",
								 query->fileName);

			revcomp_sequence (query);
			start_one_strand (target, targPositions, query, !hspsAreAdaptive,
			                  hitProc, hitProcInfo);

			// finish alignment for reverse strand hits

			if (hspsAreAdaptive) split_anchors (query->revCompFlags);
			finish_one_strand (target, targetRev, targPositions, query, NULL,
							   traceback, targCensus);


			if (hspsAreAdaptive) swap_anchors ();
			if ((hspsAreAdaptive) || (currParams->backward))
				revcomp_sequence (query);
			}

		// finish alignment for "forward" strand hits

		if (hspsAreAdaptive)
			finish_one_strand (target, targetRev, targPositions, query, NULL,
							   traceback, targCensus);

		// search for backward strand hits

		if (currParams->backward)
			{
			sthresh savedThreshold = currParams->hspThreshold;
			if (hspsAreAdaptive)
				{
				currParams->hspThreshold.t = 'S';
				currParams->hspThreshold.s = anchors->lowScore;
				}

			// search for backward strand hits

			if (currParams->verbosity >= 1)
				fprintf (stderr, "searching for matches in %s, (backward strand)\n",
								 query->fileName);

			backward_sequence (query);
			start_one_strand  (target, targPositions, query, true,
			                  hitProc, hitProcInfo);
			finish_one_strand (target, targetRev, targPositions, query, NULL,
							   traceback, targCensus);
			backward_sequence (query);

			// search for backward reverse strand hits

			if (currParams->whichStrand > 0)
				{
				if (currParams->verbosity >= 1)
					fprintf (stderr, "searching for matches in %s, (backward reverse strand)\n",
									 query->fileName);

				backward_sequence (query);
				start_one_strand  (target, targPositions, query, true,
				                   hitProc, hitProcInfo);
				finish_one_strand (target, targetRev, targPositions, query, NULL,
								   traceback, targCensus);
				}

			currParams->hspThreshold = savedThreshold;
			}
		}

	endClock = clock();

	print_m_stanza (targCensus);
	if (currParams->reportCensus)
		{
		int savedThresh = targCensus->maskThresh;
		targCensus->maskThresh = 0;
		print_census_stanza (targCensus);
		targCensus->maskThresh = savedThresh;
		}

#ifdef collect_stats
	if (currParams->gfExtend != gfexNoExtend)
		{
		print_generic (stdout, "gap_free_extensions=%jd",
		                       seed_search_hsps()
		                     + seed_search_low_scoring_hsps());
		if (!currParams->reportStats)
			print_generic (stdout, "bp_extended=%jd",
			                       seed_search_bp_extended());
		}
#endif // collect_stats

	if (currParams->reportStats)
		{
		seed_search_generic_stats   (stdout, print_generic);
		if (query->fileType == seq_type_qdna)
			quantum_generic_stats   (stdout, print_generic);
		chain_generic_stats         (stdout, print_generic);
		gapped_extend_generic_stats (stdout, print_generic);
		if (currParams->innerSeed != NULL)
			tweener_generic_stats   (stdout, print_generic);
		if (currParams->dynamicMasking > 0)
			masking_generic_stats (stdout, print_generic);
		if (currParams->inferScores)
			infer_scores_generic_stats (stdout, print_generic);
		}

	print_job_footer ();

	{
	float runTime = ((float)(endClock-startClock))/CLOCKS_PER_SEC;
	if (currParams->reportTiming)
		print_generic (stdout, "runtime=%.3f", runTime);
	lastz_set_stat (runTime, runTime);
	}

	seed_search_show_stats   (statsF);
	if (query->fileType == seq_type_qdna)
		quantum_show_stats   (statsF);
	chain_show_stats         (statsF);
	gapped_extend_show_stats (statsF);
	if (currParams->innerSeed != NULL)
		tweener_show_stats (statsF);
	if (currParams->dynamicMasking > 0)
		{
		pos_table_show_stats_after (statsF);
		masking_show_stats (statsF);
		}
	if (currParams->inferScores)
		infer_scores_show_stats (statsF);
	lastz_show_stats (statsF);

	//////////
	// clean up
	//////////

clean_up:

	memory_checkpoint ("[[* Cleanup ]]\n");

	free_if_valid        ("lz.seq1Filename",  lzParams.seq1Filename);     lzParams.seq1Filename     = NULL;
	free_sequence        (lzParams.seq1);                                 lzParams.seq1             = NULL;
	free_if_valid        ("targetRev",        targetRev);                 targetRev                 = NULL;
	free_if_valid        ("lz.seq2Filename",  lzParams.seq2Filename);     lzParams.seq2Filename     = NULL;
	free_sequence        (lzParams.seq2);                                 lzParams.seq2             = NULL;
	free_if_valid        ("lz.args",          lzParams.args);             lzParams.args             = NULL;
	free_if_valid        ("lz.scoring",       lzParams.scoring);          lzParams.scoring          = NULL;
	free_if_valid        ("lz.maskedScoring", lzParams.maskedScoring);    lzParams.maskedScoring    = NULL;
	free_position_table  (targPositions);                                 targPositions             = NULL;
	free_segment_table   (anchors);                                       anchors                   = NULL;
	free_segment_table   (savedAnchors);                                  savedAnchors              = NULL;
	free_traceback       (traceback);                                     traceback                 = NULL;
	free_traceback_rows  ();
	free_if_valid        ("targCensus",       targCensus);                targCensus                = NULL;
	free_seeds           (lzParams.hitSeed);                              lzParams.hitSeed          = NULL;
	free_seeds           (lzParams.innerSeed);                            lzParams.innerSeed        = NULL;
	fclose_if_valid      (lzParams.statsFile);                            lzParams.statsFile        = NULL;
	free_seed_hit_search ();
	free_quantum_search  ();

	free_if_valid        ("iz.inferFileName", izParams.ic.inferFileName); izParams.ic.inferFileName = NULL;
	free_seeds           (izParams.hitSeed);                              izParams.hitSeed          = NULL;
	free_if_valid        ("iz.scoring",       izParams.scoring);          izParams.scoring          = NULL;
	free_if_valid        ("iz.maskedScoring", izParams.maskedScoring);    izParams.maskedScoring    = NULL;

	// report timing stats

	dbg_timing_add (debugClockTotal);
	dbg_timing_add (debugClockQueryTotal);

	dbg_timing_report (debugClockTotal,         "total run time");
	dbg_timing_report (debugClockSeq1,          "sequence 1 I/O");
	dbg_timing_report (debugClockPosTable,      "seed position table");
	dbg_timing_report (debugClockSeq2,          "sequence 2 I/O");
	dbg_timing_report (debugClockSegTable,      "seed hit search");
	dbg_timing_report (debugClockChaining,      "chaining");
	dbg_timing_report (debugClockGappedExtend,  "gapped extension");
	dbg_timing_report (debugClockInterpolation, "interpolation");
	dbg_timing_report (debugClockOutput,        "output");
	dbg_timing_report (debugClockQueryTotal,    "total query time");

#ifdef dbgTiming
	{
	float perQuery;

	perQuery =  ((float) debugClockQueryTotal) / numQueries;
	perQuery /= clocksPerSec;

	fprintf (stderr, "%-26s %d\n",
	                 "queries:", numQueries);
	fprintf (stderr, "%-26s %.3f (%.1f per second)\n",
	                 "per query (with I/O):", perQuery, 1/perQuery);

	debugClockQueryTotal -= debugClockSeq2;
	perQuery =  ((float) debugClockQueryTotal) / numQueries;
	perQuery /= clocksPerSec;

	fprintf (stderr, "%-26s %.3f (%.1f per second)\n",
	                 "per query (w/o input):", perQuery, 1/perQuery);
	}
#endif // dbgTiming

	return 0;
	}

//----------
//
// set_up_hit_processor--
//  Set up variables that select and control the appropriate seed hit processor
//	function.
//
//----------
//
// Arguments:
//	control*		params:			Parameter set controlling the desired
//									.. search.
//	int				collectingCensus: true => caller will be collecting a
//									          .. census.
//	hitprocessor*	hitProc:		Place to return the function to call for
//									.. each hit to determine if it is 'good
//									.. enough'.
//	void**			hitProcInfo:	Place to return a value (usually a pointer
//									.. to some data) to pass thru with each
//									.. call to hitProc.
//
// Returns:
//  (nothing)
//
//----------
//
// Notes:
//	This routine is NOT reentrant, since some of the control data returned in
//	(*hitProcInfo) is stored in static variables that are private to this
//	routine.
//
//----------

void set_up_hit_processor
   (control*		params,
	int				collectingCensus,
	hitprocessor*	hitProc,
	void**			hitProcInfo)
	{
	static hitprocsimple	simpleInfo;
	static hitproctwin		twinInfo;
	hitprocinfo*			hpInfo;
	int						filtering;

	// decide which hit processor to use when we discover a seed hit

	if (params->twinMinSpan <= 0)
		{
		if ((params->gfExtend == gfexNoExtend) && (!params->gappedExtend))
			(*hitProc) = process_for_plain_hit;
		else if (params->basicHitType == hitRecover)
			(*hitProc) = process_for_recoverable_hit;
		else
			(*hitProc) = process_for_simple_hit;
		(*hitProcInfo) = (void*) &simpleInfo;
		hpInfo         = &simpleInfo.hp;
		}
	else
		{
		(*hitProc)       = process_for_twin_hit;
		(*hitProcInfo)   = (void*) &twinInfo;
		hpInfo           = &twinInfo.hp;
		twinInfo.minSpan = params->twinMinSpan;
		twinInfo.maxSpan = params->twinMaxSpan;
		}

	params->mergeAnchors = (((*hitProc) != process_for_plain_hit)
						 && ((*hitProc) != process_for_simple_hit));

	filtering =  ((currParams->minIdentity > 0)
	          ||  (currParams->maxIdentity < 1)
	          ||  (currParams->minCoverage > 0)
	          ||  (currParams->maxCoverage < 1));

	// decide how to control that hit processor

	hpInfo->reporter     = collect_hsps;
	hpInfo->reporterInfo = NULL;

	if ((anchors == NULL)
	 || ((params->hspThreshold.t =='S')
	  && (!params->gappedExtend)
	  && (!params->chain)
	  && (!params->mergeAnchors)
	  && (!collectingCensus)
	  && (!filtering)))
		hpInfo->reporter = report_hsps;

	if (params->minMatches >= 0)
		{
		hpInfo->minMatches       = params->minMatches;
		hpInfo->maxTransversions = params->maxTransversions;
		hpInfo->charToBits       = params->charToBits;
		}
	else
		hpInfo->minMatches       = -1; // (no filtering)

	if (params->gfExtend == gfexNoExtend)
		hpInfo->gfExtend         = gfexNoExtend;
	else if (params->gfExtend == gfexExact)
		{
		hpInfo->gfExtend         = gfexExact;
		hpInfo->seq1             = params->seq1;
		hpInfo->seq2             = params->seq2;
		hpInfo->hspThreshold     = params->hspThreshold;
		hpInfo->anchors          = &anchors;
		seed_search_set_stat(isHspSearch,true);
		}
	else // if (params->gfExtend == gfexXDrop)
		{
		hpInfo->gfExtend         = gfexXDrop;
		hpInfo->seq1             = params->seq1;
		hpInfo->seq2             = params->seq2;
		hpInfo->scoring          = params->maskedScoring;
		hpInfo->xDrop            = params->xDrop;
		hpInfo->hspThreshold     = params->hspThreshold;
		hpInfo->hspZeroThreshold = (params->hspThreshold.t !='S')? 0
		                         : (params->hspThreshold.s >  0 )? params->hspThreshold.s
		                                                         : 0;
		hpInfo->anchors          = &anchors;
		hpInfo->entropicHsp      = params->entropicHsp;
		hpInfo->reportEntropy    = params->reportEntropy;
		seed_search_set_stat(isHspSearch,true);

		if (infer_scores_dbgShowIdentity)
			{
			printf ("hit_processor xDrop = "      scoreFmtSimple "\n",
			        hpInfo->xDrop);
			printf ("hit_processor hspThreshold = %s\n",
			        score_thresh_to_string (&hpInfo->hspThreshold));
			}
		}
	}

//----------
//
// start_one_strand--
//  Start alignment upon one query strand.
//
//----------
//
// Arguments:
//	seq*			target:			The sequence being searched.
//	postable*		targPositions:	A table of positions of words in target.
//	seq*			query:			The sequence(s) being searched for.
//	int				emptyAnchors:	true => clear the anchors table before
//									        .. starting.
//	hitprocessor	hitProc:		Function to call for each hit to determine
//									.. if it is 'good enough'.
//	void*			hitProcInfo:	A value (usually a pointer to some data) to
//									.. pass thru with each call to hitProc.
//
// Returns:
//  (nothing)
//
//----------

static int headerPrinted;	// false => we have yet to print a header for this
							//          .. strand-to-strand alignment

void start_one_strand
   (seq*			target,
	postable*		targPositions,
	seq*			query,
	int				emptyAnchors,
	hitprocessor	hitProc,
	void*			hitProcInfo)
	{
	// find seed hits;  depending on hitProc, we will may extend these to HSPs;
	// and depending on format parameters and hitReporter, we will either report
	// these directly to the output, or we will collect them in anchors[]

	dbg_timing_sub (debugClockSegTable);

	if ((emptyAnchors) && (anchors != NULL))
		empty_segment_table (anchors);

	headerPrinted = false;
	if (query->fileType == seq_type_qdna)
		quantum_seed_hit_search (target, targPositions,
		                         query, 0, query->len, currParams->coding,
		                         currParams->charToBits, currParams->hitSeed,
		                         currParams->maskedScoring, currParams->ballScore,
		                         hitProc, hitProcInfo);
	else
		seed_hit_search (target, targPositions,
		                 query, 0, query->len, currParams->selfCompare,
		                 currParams->charToBits, currParams->hitSeed,
		                 hitProc, hitProcInfo);

	dbg_timing_add (debugClockSegTable);
	}

//----------
//
// finish_one_strand--
//  Finish alignment upon one query strand.
//
//----------
//
// Arguments:
//	seq*			target:			The sequence being searched.
//	postable*		targPositions:	A table of positions of words in target.
//	u8*				targetRev:		The reverse (NOT reverse complement) of the
//									.. target sequence, as a zero-terminated
//									.. string;  this may be NULL if the caller
//									.. doesn't need/want to supply it.  It is
//									.. only needed if we will be doing a gapped
//									.. extension.
//	seq*			query:			The sequence(s) being searched for.
//	u8*				queryRev:		The reverse (NOT reverse complement) of the
//									.. query sequence (analagous to targetRev)
//	tback*			traceback:		Memory in which to track gapped alignment
//									.. traceback.
//	census*			targCensus:		Census array for target sequence.  If this
//									.. is non-NULL, we count how many times each
//									.. target base is aligned.  This information
//									.. is used to mask positions in the target
//									.. sequence if currParams->dynamicMasking > 0.
//
// Returns:
//  (nothing)
//
//----------

void finish_one_strand
   (seq*			target,
	u8*				_targetRev,
	postable*		targPositions,
	seq*			query,
	u8*				_queryRev,
	tback*			traceback,
	census*			targCensus)
	{
	u8*				targetRev = _targetRev;
	u8*				queryRev  = _queryRev;
	alignel*		alignList = NULL;

	if ((anchors != NULL)				// merging may be necessary because
	 && (currParams->mergeAnchors))		// .. the diag hash technique used in
		merge_segments (anchors);		// .. the seed search may result in
										// .. duplicate or overlapping HSPs

	// filter HSPs by identity and/or coverage, if desired

	if ((anchors != NULL)
	 && (!currParams->gappedExtend))
		{
		if ((currParams->minIdentity > 0) || (currParams->maxIdentity < 1))
			filter_segments_by_identity (target, query, anchors,
			                             currParams->minIdentity,
			                             currParams->maxIdentity);

		if ((currParams->minCoverage > 0) || (currParams->maxCoverage < 1))
			filter_segments_by_coverage (target, query, anchors,
			                             currParams->minCoverage,
			                             currParams->maxCoverage);
		}

	// reduce the set of HSPs to the best syntenic subset

	if (currParams->chain)
		{
		dbg_timing_sub  (debugClockChaining);
		reduce_to_chain (anchors, currParams->chainDiag, currParams->chainAnti,
		                 chainScale, chain_connect_penalty);
		sort_segments   (anchors, qSegmentsByPos1);
		dbg_timing_add  (debugClockChaining);
		}

	// report the set of HSPs if we don't plan to do gapped extension

	if ((anchors != NULL) && (!currParams->gappedExtend))
		{
		u32			ix;
		segment*	seg;

		for (ix=0,seg=anchors->seg ; ix<anchors->len ; ix++,seg++)
			print_match (seg->pos1, seg->pos2, seg->length, seg->s);
		}

	// if we don't plan to do gapped extension, perform dynamic masking;  note
	// that if currParams->dynamicMasking == 0, we don't actually mask, we just
	// count for the census;  further note that we don't mask the reverse
	// sequence unless the caller provided it

	if ((targCensus != NULL) && (anchors != NULL) && (!currParams->gappedExtend))
		{
		unspos numMasked;
		numMasked = census_mask_segments
		               (anchors, target->v, _targetRev, targCensus,
		                remove_interval_seeds, targPositions);
		print_x_stanza (numMasked);
		}

	// extend the HSPs to gapped alignments

	if (currParams->gappedExtend)
		{
		dbg_timing_sub (debugClockGappedExtend);
		if (targetRev == NULL)
			targetRev = (u8*) reverse_of_string ((char*) target->v, target->len);
		if (queryRev == NULL)
			queryRev =  (u8*) reverse_of_string ((char*) query->v,  query->len);

		reduce_to_points (target, query, currParams->scoring, anchors);
		if ((anchors != NULL) && (dbgShowAnchors))
			dump_segments (stderr, anchors,
						   ((target->revCompFlags & rcf_rev) != 0)?"-":"+",
						   ((query->revCompFlags  & rcf_rev) != 0)?"-":"+");
		alignList = gapped_extend (target, targetRev, query, queryRev,
		                           currParams->selfCompare,
		                           currParams->scoring, anchors, traceback,
		                           currParams->yDrop, currParams->gappedThreshold);
		dbg_timing_add (debugClockGappedExtend);
		}

	// filter gapped alignments by identity and/or coverage, if desired

	if (alignList != NULL)
		{
		if ((currParams->minIdentity > 0) || (currParams->maxIdentity < 1))
			alignList = filter_aligns_by_identity (target, query, alignList,
			                                       currParams->minIdentity,
			                                       currParams->maxIdentity);
		if ((currParams->minCoverage > 0) || (currParams->maxCoverage < 1))
			alignList = filter_aligns_by_coverage (target, query, alignList,
			                                       currParams->minCoverage,
			                                       currParams->maxCoverage);
		}

	// interpolate between the gapped alignments
	// nota bene:  we use yDrop for both drop parameters to match blastz

	if ((alignList != NULL) && (currParams->innerThreshold > 0))
		{
		dbg_timing_sub (debugClockInterpolation);
		alignList = tweener_interpolate
		              (alignList, target, query, currParams->selfCompare,
		               currParams->charToBits, currParams->innerSeed,
		               currParams->scoring, currParams->maskedScoring, traceback,
		               currParams->yDrop, currParams->yDrop, currParams->innerThreshold,
				       currParams->chainDiag, currParams->chainAnti,
				       chainScale, chain_connect_penalty, currParams->innerWindow);
		dbg_timing_add (debugClockInterpolation);
		}

	// print the gapped alignments

	if (alignList != NULL)
		{
		dbg_timing_sub (debugClockOutput);
		print_align_list (alignList);
		dbg_timing_add (debugClockOutput);
		}

	// perform dynamic masking;  note that if currParams->dynamicMasking == 0, we
	// don't actually mask, we just count for the census;  further note that
	// we don't mask the reverse sequence unless the caller provided it

	if ((targCensus != NULL) && (alignList != NULL))
		{
		unspos numMasked;
		numMasked = census_mask_aligns
		               (alignList, target->v, _targetRev, targCensus,
		                remove_interval_seeds, targPositions);
		print_x_stanza (numMasked);
		}

	// cleanup

	if (alignList != NULL)
		free_align_list  (alignList);

	if ((_targetRev == NULL) && (targetRev != NULL))
		free_if_valid ("process_one_strand (targetRev)", targetRev);
	if ((_queryRev  == NULL) && (queryRev  != NULL))
		free_if_valid ("process_one_strand (queryRev)", queryRev);
	}

//----------
//
// split_anchors, swap_anchors--
//	Select single-strand anchors from a list containing anhors for both strands.
//
// Under certain configurations (e.g. for adaptive-K), collect_hsps collects
// HSPs from both strands into a single table (anchors).  Split_anchors removes
// the forward stand HSPs and moves them to a second table (savedAnchors).
// Swap_anchors switches the tables so that the forward strand HSPs are in the
// main table (anchors).
//
//----------
//
// Arguments:
//	int		id:		The id of segments on the reverse strand.
//	(none)
//
// Returns:
//  (nothing)
//
//----------

void split_anchors (int id)
	{
	if (savedAnchors == NULL)
		savedAnchors = new_segment_table (numDefaultAnchors,
		                                  anchors->coverageLimit);
	else
		empty_segment_table (savedAnchors);

	split_segment_table (anchors, id, &savedAnchors);

	//printf ("\nanchors:\n");
	//dump_segments (stdout, anchors, NULL, NULL);
	//printf ("\nleftovers:\n");
	//dump_segments (stdout, savedAnchors, NULL, NULL);
	//printf ("\n");
	}


void swap_anchors (void)
	{ segtable* a = savedAnchors;  savedAnchors = anchors;  anchors = a; }

//----------
// [[-- a chain connection penalty function --]]
//
// chain_connect_penalty--
//	Compute penalty for connecting two segments in the chain.
//
// Arguments and Return value: (see chain.h)
//
// Note bene:
//	x = pos1, y = pos2, diag = x-y
//	diag increases toward southeast
//
//----------

static score chain_connect_penalty
   (segment*	seg1,
	segment*	seg2,
	int			scale)
	{
	unspos		xEnd, yEnd;
	sgnpos		diag1, diag2, diagDiff;
	sgnpos		numSubs;	// number of substitutions needed to get from end
	double		penalty;	// .. of segment 1 to beginning of segment 2

	if ((seg2->pos1 <= seg1->pos1) || (seg2->pos2 <= seg1->pos2))
		suicide ("HSPs improperly ordered for chaining");

	xEnd  = seg1->pos1 + seg1->length - 1;
	yEnd  = seg1->pos2 + seg1->length - 1;

	diag1 = diagNumber (seg1->pos1, seg1->pos2);
	diag2 = diagNumber (seg2->pos1, seg2->pos2);

	diagDiff = diag2 - diag1;
	if (diagDiff >= 0)
		{					// segment 1's diagonal is above segment 2's
		numSubs = ((sgnpos) seg2->pos2) - ((sgnpos) yEnd) - 1;
		}
	else
		{					// segment 1's diagonal is below segment 2's
		numSubs  = ((sgnpos) seg2->pos1) - ((sgnpos) xEnd) - 1;
		diagDiff = -diagDiff;
		}

	// nota bene: penalty is declared as double to allow it to overflow the
	//            regular score type;  after we compute the penalty we clip it
	//            to worst penalty as we return it

	penalty = diagDiff * currParams->chainDiag;
	if (numSubs >= 0)
		penalty +=   numSubs  * currParams->chainAnti;
	else
		penalty += (-numSubs) * scale * currParams->scoring->sub['A']['A'];

	if (penalty > bestPossibleScore) return bestPossibleScore;
	                            else return (score) penalty;
	}

//----------
// [[-- a report_census_intervals callback function --]]
//
// remove_interval_seeds--
//	Remove seeds from the target sequence position table that are about to be
//	rendered meaningless by dynamic masking.
//
//----------
//
// Arguments:
//	unspos b, e:	The interval, in the target sequence, that is about to be
//					.. masked.  Origin-1, inclusive.
//	void* info:		(really postable*) The position table (targPositions).
//
// Returns:
//  (nothing)
//
//----------
//
// Notes:
//	(1)	Intervals given to this routine are origin-1 inclusive, while the
//		intervals it passes along to mask_seed_position_table are origin-0
//		end-exclusive.  Further, we have to expand the interval on both ends
//		by the length of the seed (minus 1).
//
//		For example, suppose the input interval is 20..40 (11 bp) and the seed
//		length is 10.  In the diagram below, * is a base in the input interval,
//		x is a base in the expanded interval, and o is any other base.  Numbers
//		on the top are origin-1; numbers on the bottom are origin-0.  The output
//		interval is 10..49.
//
//		    1        11       20                  40       49
//			v         v        v                   v        v
//			ooooooooooxxxxxxxxx*********************xxxxxxxxxoooooooo ...
//			^         ^        ^                   ^        ^
//			0        10       19                  39       48
//
//		The reason for expansion is that we want to eliminate any seed that
//		contains any * in the diagram.  The * can occur at any position in the
//		10 bp seed, so we must expand by 9 bp on each end.
//
//----------

static void remove_interval_seeds (unspos b, unspos e, void* info)
	{
	postable*	pt = (postable*) info;
	seq*		target  = currParams->seq1;
	seed*		hitSeed = currParams->hitSeed;
	u32			seedLen = (unsigned) hitSeed->length;
	const s8*	charToBits = currParams->charToBits;

	// adjust the interval endpoints to account for the seed length

	if (b < seedLen) b =  1;
	            else b -= seedLen - 1;

	if (e >= target->len - (seedLen-1)) e =  target->len;
	                               else e += seedLen - 1;

	// remove masked seeds from the table

	mask_seed_position_table (pt, target, b-1, e, charToBits, hitSeed);
	}

//----------
// [[-- a seed hit reporter function --]]
//
// report_hsps--
//	Report a seed hit or HSP (i.e. just write it to output).
//
// Arguments and Return value: (see seed_search.h)
//
//----------

static u32 report_hsps
   (arg_dont_complain(void* info),
	unspos	pos1,
	unspos	pos2,
	unspos	length,
	score	s)
	{
	unspos	s1, s2;

	// report this hit/HSP

	print_match (pos1-length, pos2-length, length, s);

	if (dbgShowHsps)
		{
		fprintf (stderr, "\n");
		dump_aligned_nucleotides (stderr,
		                          currParams->seq1, pos1-length,
		                          currParams->seq2, pos2-length,
		                          length);
		}

	if (!currParams->mirror)
		return 1;

	// report the mirror of this hit/HSP

	if (currParams->seq1->revCompFlags == currParams->seq2->revCompFlags)
		{
		s1 = pos1;
		s2 = pos2;
		}
	else
		{
		s1 = (currParams->seq1->len) - pos1 + length;
		s2 = (currParams->seq2->len) - pos2 + length;
		if ((s2 == pos1) && (s1 == pos2)) return 1;
		}

	print_match (s2-length, s1-length, length, s);

	if (dbgShowHsps)
		{
		fprintf (stderr, "\n");
		dump_aligned_nucleotides (stderr,
		                          currParams->seq1, s2-length,
		                          currParams->seq2, s1-length,
		                          length);
		}

	return 2;
	}

//----------
// [[-- a seed hit reporter function --]]
//
// collect_hsps--
//	Collect a seed hit or HSP.
//
// Arguments and Return value: (see seed_search.h)
//
//----------

static u32 collect_hsps
   (arg_dont_complain(void* info),
	unspos	pos1,
	unspos	pos2,
	unspos	length,
	score	s)
	{
	unspos	s1, s2;

	// add this hit/HSP to the list of anchors;  note that we use the strand
	// as the id field, so that if we happen to be collecting segments from
	// both strands, we can separate them later

	anchors = add_segment (anchors, pos1-length, pos2-length, length, s,
	                       /* id */ currParams->seq2->revCompFlags);

	if (dbgShowHsps)
		{
		fprintf (stderr, "\n");
		dump_aligned_nucleotides
			 (stderr, currParams->seq1, pos1-length, currParams->seq2, pos2-length, length);
		}

	if (!currParams->mirror)
		return 1;

	// add the mirror of this hit/HSP to the list of anchors

	if (currParams->seq1->revCompFlags == currParams->seq2->revCompFlags)
		{
		s1 = pos1;
		s2 = pos2;
		}
	else
		{
		s1 = (currParams->seq1->len) - pos1 + length;
		s2 = (currParams->seq2->len) - pos2 + length;
		if ((s2 == pos1) && (s1 == pos2)) return 1;
		}

	anchors = add_segment (anchors, s2-length, s1-length, length, s,
	                       /* id */ currParams->seq2->revCompFlags);

	if (dbgShowHsps)
		{
		fprintf (stderr, "\n");
		dump_aligned_nucleotides
			 (stderr, currParams->seq1, s2-length, currParams->seq2, s1-length, length);
		}

	return 2;
	}

//----------
//
// parse_options--
//  Parse command line options.
//
//----------
//
// Arguments:
//	argc, argv:	(as per main)
//	control*	lzParams:	Control data to fill in for the primary alignment.
//	control*	izParams:	Control data to fill in for inference alignments.
//
// Returns:
//  (nothing;  failure causes program fatality)
//
//----------

// pre-canned expansion arguments

typedef struct exparg
	{
	char*	argName;
	char*	expansion;
	} exparg;

exparg expanders[] =
	{
// old expansions:
//	{ "--yasra98",      "T=2 Z=20 --match=1,6 O=6 E=1 Y=20 K=22 L=30 --identity=98..100" },
//	{ "--yasra95",      "T=2 Z=20 --match=1,5 O=8 E=1 Y=40 K=40 L=40 --identity=95..100" },
//	{ "--yasra90",      "T=2 Z=20 --match=1,5 O=6 E=1 Y=40 K=40 L=40 --identity=90..100" },
//	{ "--yasra85",      "T=2      --match=1,2 O=4 E=1 Y=20 K=22 L=30 --identity=85..100" },
//	{ "--yasra75",      "T=2      --match=1,1 O=3 E=1 Y=20 K=22 L=30 --identity=75..100" }
//	{ "--yasra95short", "T=2 --match=1,7 K=10 --nogapped --noentropy --identity=95..100" },
//	{ "--yasra85short", "T=2 --match=1,2 K=11 --nogapped --noentropy --identity=85..100" }
	{ "--yasra98",      "T=2 Z=20 --match=1,6 O=8 E=1 Y=20 K=22 L=30 --identity=98..100" },
	{ "--yasra95",      "T=2 Z=20 --match=1,5 O=8 E=1 Y=20 K=22 L=30 --identity=95..100" },
	{ "--yasra90",      "T=2 Z=20 --match=1,5 O=6 E=1 Y=20 K=22 L=30 --identity=90..100" },
	{ "--yasra85",      "T=2      --match=1,2 O=4 E=1 Y=20 K=22 L=30 --identity=85..100" },
	{ "--yasra75",      "T=2      --match=1,1 O=3 E=1 Y=20 K=22 L=30 --identity=75..100" },
	{ "--yasra95short", "T=2      --match=1,7 O=6 E=1 Y=14 K=10 L=14 --identity=95..100" },
	{ "--yasra85short", "T=2      --match=1,3 O=4 E=1 Y=14 K=11 L=14 --identity=85..100" }
	};
#define numExpanders ((int)(sizeof(expanders)/sizeof(exparg)))


static void usage (void);
static void usage (void)
	{
	fprintf (helpout, "%s-- search for seed hits in DNA sequences\n", programName);
	fprintf (helpout, "  (version %s.%s.%s built %s",
	                  programVersionMajor, programVersionMinor, programVersionSubMinor, programRevisionDate);
	if      (scoreType == 'F') fprintf (helpout, ", floating point scores");
	else if (scoreType == 'D') fprintf (helpout, ", double floating point scores");
	fprintf (helpout, ")\n");
	fprintf (helpout, "usage: %s target [query] [options]\n", programName);

	fprintf (helpout, "  (common options;  use --help for a more extensive list)\n");
	fprintf (helpout, "  target, query          specifiers or files, containing sequences to align\n");
	fprintf (helpout, "                         (use --help=files for more details)\n");

	fprintf (helpout, "  --seed=<pattern>       set seed pattern (12of19, 14of22, or general pattern)\n");
	fprintf (helpout, "                         (default is %s)\n",
	         defaultSeedString);
	fprintf (helpout, "  --[no]transition       allow (or don't) one transition in a seed hit\n");
	fprintf (helpout, "                         (by default %s)\n",
	         (defaultParams.withTrans == 0)? "the seed must match as is" :
	         (defaultParams.withTrans == 1)? "a transition is allowed"
	                                       : "two transitions are allowed");

	fprintf (helpout, "  --[no]chain            perform chaining\n");
	fprintf (helpout, "                         (by default %s)\n",
	         (!defaultParams.chain)?  "no chaining is performed"
	                                : "chaining is performed");

	fprintf (helpout, "  --[no]gapped           perform gapped alignment (instead of gap-free)\n");
	fprintf (helpout, "                         (by default %s)\n",
	         (!defaultParams.gappedExtend)? "gapped alignment is not performed"
	                                      : "gapped alignment is performed");

	fprintf (helpout, "  --step=<length>        set step length (default is %u)\n",
	         defaultParams.step);

	fprintf (helpout, "  --both                 search both strands\n");
	fprintf (helpout, "  --plus                 search + strand only (matching strand of query spec)\n");
	fprintf (helpout, "                         (by default %s)\n",
	         (defaultParams.whichStrand == 0)? "only + strand is searched" :
	         (defaultParams.whichStrand <  0)? "only - strand is searched"
	                                         : "both strands are searched");

	fprintf (helpout, "  --scores=<file>        read substitution and gap scores from a file\n");
	fprintf (helpout, "  --xdrop=<score>        set x-drop threshold (default is 10sub[A][A])\n");
	fprintf (helpout, "  --ydrop=<score>        set y-drop threshold (default is open+300extend)\n");

	fprintf (helpout, "  --infer[=<control>]    infer scores from the sequences, then use them\n");
	fprintf (helpout, "                         all inference options are read from the control file\n");

	fprintf (helpout, "  --hspthresh=<score>    set threshold for high scoring pairs (default is %s)\n",
	         score_thresh_to_string (&defaultParams.hspThreshold));
	fprintf (helpout, "                         ungapped extensions scoring lower are discarded\n");
	fprintf (helpout, "                         <score> can also be a percentage or base count\n");

	fprintf (helpout, "  --gappedthresh=<score> set threshold for gapped alignments\n");
	fprintf (helpout, "                         gapped extensions scoring lower are discarded\n");
	fprintf (helpout, "                         <score> can also be a percentage or base count\n");
	fprintf (helpout, "                         (default is to use same value as --hspthresh)\n");

	fprintf (helpout, "  --help                 list all options\n");
	fprintf (helpout, "  --help=files           list information about file specifiers\n");
	fprintf (helpout, "  --help=short[cuts]     list blastz-compatible shortcuts\n");

	exit (1);
	}


static void all_options (void);
static void all_options (void)
	{
	int ix;

	fprintf (helpout, "  target[[s..e]][-]      spec/file containing target sequence (fasta or nib)\n");
	fprintf (helpout, "                         [s..e] defines a subrange of the file\n");
	fprintf (helpout, "                         - indicates reverse-complement\n");
	fprintf (helpout, "                         (use --help=files for more details)\n");
	fprintf (helpout, "  query[[s..e]][-]       spec/file containing query sequences (fasta or nib)\n");
	fprintf (helpout, "                         if absent, queries come from stdin (unless they\n");
	fprintf (helpout, "                         aren't needed, as for --self or --tableonly)\n");
	fprintf (helpout, "                         (use --help=files for more details)\n");
	fprintf (helpout, "  --self                 the target sequence is also the query\n");
	fprintf (helpout, "  --quantum              the query sequence contains quantum DNA\n");

	fprintf (helpout, "  --seed=match(<length>) use a word with no gaps instead of a seed pattern\n");
	fprintf (helpout, "  --seed=half(<length>)  use space-free half-weight word instead of seed pattern\n");
	fprintf (helpout, "  --[no]trans[ition][=2] allow one or two transitions in a seed hit\n");
	fprintf (helpout, "                         (by default %s)\n",
	         (defaultParams.withTrans == 0)? "the seed must match as is" :
	         (defaultParams.withTrans == 1)? "a transition is allowed"
	                                       : "two transitions are allowed");

	fprintf (helpout, "  --word=<bits>          set max bits for word hash;  use this to trade time for\n");
	fprintf (helpout, "                         memory, eliminating thrashing for heavy seeds\n");
	fprintf (helpout, "                         (default is %d bits)\n",
	         defaultParams.maxIndexBits);

	fprintf (helpout, "  --[no]filter=[<T>:]<M> filter half-weight seed hits, requiring at least M\n");
	fprintf (helpout, "                         matches and allowing no more than T transversions\n");
	if (defaultParams.minMatches < 0)
		fprintf (helpout, "                         (default is no filtering)\n");
	else if (defaultParams.maxTransversions == 0)
		fprintf (helpout, "                         (default is to require %d matches)\n",
		         defaultParams.minMatches);
	else
		fprintf (helpout, "                         (default is %d matches/%d transversions)\n",
		         defaultParams.minMatches,defaultParams.maxTransversions);

	fprintf (helpout, "  --notwins              require just one seed hit\n");
	fprintf (helpout, "  --twins=[<min>:]<maxgap> require two nearby seed hits on the same diagonal\n");
	if (defaultTwinsYes)
		fprintf (helpout, "                         (default is twins with %d:%d bp gap)\n",
		         defaultTwinMinGap,defaultTwinMaxGap);
	else
		fprintf (helpout, "                         (default is twins aren't required)\n");

	fprintf (helpout, "  --recoverhits          recover hash-collision seed hits\n");
	if (defaultParams.basicHitType == hitRecover)
		fprintf (helpout, "                         (default is to recover seed hits)\n");
	else
		fprintf (helpout, "                         (default is not to recover seed hits)\n");

	fprintf (helpout, "  --step=<length>        set step length (default is %u)\n",
	         defaultParams.step);

	fprintf (helpout, "  --both[strands]        search both strands\n");
	fprintf (helpout, "  --plus[strand]         search + strand only (matching strand of query spec)\n");
	fprintf (helpout, "  --minus[strand]        search - strand only (opposite strand of query spec)\n");
	fprintf (helpout, "                         (by default %s)\n",
	         (defaultParams.whichStrand == 0)? "only + strand is searched" :
	         (defaultParams.whichStrand <  0)? "only - strand is searched"
	                                         : "both strands are searched");

	fprintf (helpout, "  --ambiguousn           treat N as an ambiguous nucleotide\n");
	fprintf (helpout, "                         (by default N is treated as a sequence splicing character)\n");

	fprintf (helpout, "  --[no]gfextend         perform gap-free extension of seed hits to HSPs\n");
	fprintf (helpout, "                         (by default %s)\n",
	           (!defaultParams.gfExtend == gfexNoExtend)?  "no extension is performed"
	         : (!defaultParams.gfExtend == gfexExact)?     "exact match extension is performed"
	                                                     : "extension is performed");

	fprintf (helpout, "  --[no]chain            perform chaining\n");
	fprintf (helpout, "  --chain=<diag,anti>    perform chaining with given penalties for diagonal and\n");
	fprintf (helpout, "                         anti-diagonal\n");
	fprintf (helpout, "                         (by default %s)\n",
	         (!defaultParams.chain)?  "no chaining is performed"
	                                : "chaining is performed");

	fprintf (helpout, "  --[no]gapped           perform gapped alignment (instead of gap-free)\n");
	fprintf (helpout, "                         (by default %s)\n",
	         (!defaultParams.gappedExtend)? "gapped alignment is not performed"
	                                      : "gapped alignment is performed");

	fprintf (helpout, "  --score[s]=<file>      read substitution scores from a file\n");
	fprintf (helpout, "                         (default is HOXD70)\n");
	fprintf (helpout, "  --unitscore[s]         scores are +1/-1 for match/mismatch\n");
	fprintf (helpout, "  --gap=<[open,]extend>  set gap open and extend penalties (default is " scoreFmtSimple "," scoreFmtSimple ")\n",
	                 HOXD70_open, HOXD70_extend);
	fprintf (helpout, "  --xdrop=<score>        set x-drop threshold (default is 10*sub[A][A])\n");
	fprintf (helpout, "  --ydrop=<score>        set y-drop threshold (default is open+300extend)\n");

	fprintf (helpout, "  --infer[=<control>]    infer scores from the sequences, then use them\n");
	fprintf (helpout, "  --inferonly[=<control>] infer scores, but don't use them (requires --infscores)\n");
	fprintf (helpout, "                         all inference options are read from the control file\n");
	fprintf (helpout, "  --infscores[=<file>]   write inferred scores to a file\n");

	fprintf (helpout, "  --hspthresh=<score>    set threshold for high scoring pairs (default is %s)\n",
	         score_thresh_to_string (&defaultParams.hspThreshold));
	fprintf (helpout, "                         ungapped extensions scoring lower are discarded\n");
	fprintf (helpout, "                         <score> can also be a percentage or base count\n");

	fprintf (helpout, "  --exact=<length>       set threshold for exact matches\n");
	fprintf (helpout, "                         if specified, exact matches are found rather than high\n");
	fprintf (helpout, "                         scoring pairs (replaces --hspthresh)\n");

	fprintf (helpout, "  --inner=<score>        set threshold for HSPs during interpolation\n");
	if (defaultParams.innerThreshold <= 0)
		fprintf (helpout, "                         (default is no interpolation)\n");
	else
		fprintf (helpout, "                         (default is " scoreFmtSimple ")\n",
		         defaultParams.innerThreshold);

	fprintf (helpout, "  --gappedthresh=<score> set threshold for gapped alignments\n");
	fprintf (helpout, "                         gapped extensions scoring lower are discarded\n");
	fprintf (helpout, "                         <score> can also be a percentage or base count\n");
	fprintf (helpout, "                         (default is to use same value as --hspthresh)\n");

	fprintf (helpout, "  --ball=<score>         set minimum score required of words 'in' a quantum ball\n");

	fprintf (helpout, "  --[no]entropy          involve entropy in filtering high scoring pairs\n");
	fprintf (helpout, "                         (default is \"%s\")\n",
	         (defaultParams.entropicHsp)? "entropy"
	                                    : "noentropy");

	fprintf (helpout, "  --[no]mirror           report/use mirror image of all gap-free alignments\n");
	fprintf (helpout, "                         (default is \"mirror\" for self-alignments only)\n");

	fprintf (helpout, "  --traceback=<bytes>    space for trace-back information\n");
	fprintf (helpout, "                         (default is %s)\n",
	         unitize(defaultParams.tracebackMem));

	fprintf (helpout, "  --masking=<count>      mask any position in target hit this many times\n");
	fprintf (helpout, "                         zero indicates no masking\n");
	if (defaultParams.dynamicMasking <= 0)
		fprintf (helpout, "                         (default is no masking)\n");
	else
		fprintf (helpout, "                         (default is %d)\n",
		         defaultParams.dynamicMasking);

	fprintf (helpout, "  --[no]census           count/report how many times each target base aligns\n");
	fprintf (helpout, "                         (default is %s)\n",
	         (defaultParams.reportCensus)? "to report census"
	                                     : "to not report census");

	fprintf (helpout, "  --identity=<min>[..<max>] filter alignments by percent identity\n");
	fprintf (helpout, "                         0<=min<=max<=100;  blocks (or HSPs) outside min..max\n");
	fprintf (helpout, "                         are discarded\n");
	fprintf (helpout, "                         (default is no identity filtering)\n");

	fprintf (helpout, "  --coverage=<min>[..<max>] filter alignments by percentage pf query covered\n");
	fprintf (helpout, "                         0<=min<=max<=100;  blocks (or HSPs) outside min..max\n");
	fprintf (helpout, "                         are discarded\n");
	fprintf (helpout, "                         (default is no query coverage filtering)\n");

//	fprintf (helpout, "  --[no]laj              backward compatibility for laj\n");
//	fprintf (helpout, "                         (default is %s)\n",
//	         (defaultParams.lajCompatible)? "to be backward compatible"
//	                                      : "not to bother with backward compatibility");

	fprintf (helpout, "  --code=<file>          give quantum code for query sequence (only for display)\n");

	fprintf (helpout, "  --format=<type>        specify output format; one of lav, axt, maf, text,\n");
	fprintf (helpout, "                         lav+text, gfa, identity, infstats(<pct id range>)\n");
	fprintf (helpout, "                         (by default output is %s)\n",
	         formatNames[defaultParams.outputFormat]);

	fprintf (helpout, "  --verbosity=<level>    set info level (0 is minimum, 10 is everything)\n");
	fprintf (helpout, "                         (default is %d)\n",
	         defaultParams.verbosity);

	fprintf (helpout, "  --[no]runtime          report runtime in the output file\n");
	fprintf (helpout, "                         (default is %s)\n",
	         (defaultParams.reportTiming)? "to report runtime"
	                                     : "to not report runtime");

	fprintf (helpout, "  --tableonly[=count]    just produce the target position table, don't\n");
	fprintf (helpout, "                         search for seeds\n");

	fprintf (helpout, "  --[no]stats[=<file>]   show search statistics (or don't)\n");
#ifdef collect_stats
	fprintf (helpout, "                         (default is %s)\n",
	         (defaultParams.showStats)? "to show shats"
	                                  : "to not show stats");
#else
	fprintf (helpout, "                         (not available in this build)\n");
#endif

	fprintf (helpout, "  --version              report the program version and quit\n");
	fprintf (helpout, "  --help                 list all options\n");
	fprintf (helpout, "  --help=files           list information about file specifiers\n");
	fprintf (helpout, "  --help=short[cuts]     list blastz-compatible shortcuts\n");

	for (ix=0 ; ix<numExpanders ; ix++)
		fprintf (helpout, "  %-22s (%s)\n",
		                  expanders[ix].argName, expanders[ix].expansion);

	exit (1);
	}

static void file_options (void);
static void file_options (void)
	{
	fprintf (helpout, "usage: %s target [query] [options]\n", programName);
	fprintf (helpout, "\n");
	fprintf (helpout, "The general form of both target and query specifiers is\n");
	fprintf (helpout, "    [nickname::]filename[/selectname][[actions]][-]\n");
	fprintf (helpout, "\n");
	fprintf (helpout, "The filename field is required; all other fields are optional.\n");
	fprintf (helpout, "\n");
	fprintf (helpout, "Fields:\n");
	fprintf (helpout, "  nickname         name to use for this sequence in any output files\n");
	fprintf (helpout, "  filename         file (or path) to fasta, nib or 2bit file\n");
	fprintf (helpout, "  selectname       read only a single sequence from the file\n");
	fprintf (helpout, "                   (only valid for 2bit)\n");
	fprintf (helpout, "  actions          list of pre-processing actions;  enclosed in square brackets\n");
	fprintf (helpout, "                   and comma-separated;  see list of actions below\n");
	fprintf (helpout, "  - (minus sign)   use reverse complement of the sequence\n");
	fprintf (helpout, "                   (equivalent to the revcomp action listed below)\n");
	fprintf (helpout, "\n");
	fprintf (helpout, "Supported actions:\n");
	fprintf (helpout, "  unmask           convert any lowercase bases to uppercase\n");
	fprintf (helpout, "  revcomp          reverse complement\n");
	fprintf (helpout, "  multi[ple]       file's sequences are internally treated as a single sequence\n");
	fprintf (helpout, "  @namesfile       process only the sequences listed in namesfile\n");
	fprintf (helpout, "                   (only valid for fasta and 2bit)\n");
	fprintf (helpout, "  name:al[pha]num  pull a short name from sequence header, alphanumeric only\n");
	fprintf (helpout, "  name=tag         pull a short name from sequence header, starting from tag\n");
	fprintf (helpout, "                   (only valid for fasta)\n");
	fprintf (helpout, "  <subrange>       only process a subrange of the file (see below)\n");
	fprintf (helpout, "\n");
	fprintf (helpout, "Subranges:\n");
	fprintf (helpout, "  start,end        process from start thru end, inclusive\n");
	fprintf (helpout, "  [start]..[end]   absent fields use 1 or sequence length\n");
	fprintf (helpout, "  start#length     same as start..start+length-1\n");
	fprintf (helpout, "  (subrange indices begin with 1 and are inclusive)\n");

	exit (1);
	}

static void shortcuts (void);
static void shortcuts (void)
	{
	scoreset* hoxScoring = NULL;

	fprintf (helpout, "%54s%s\n", "", "[defaults]");
	fprintf (helpout, "  B=0          same as --plusstrand%s\n",
	                 (defaultParams.whichStrand == 0)? "                   [B=0]" : "");
	fprintf (helpout, "  B=2          same as --bothstrands\%s\n",
	                 (defaultParams.whichStrand > 0)? "                  [B=2]" : "");
	fprintf (helpout, "  B=-1         same as --minusstrand%s\n",
	                 (defaultParams.whichStrand < -1)? "                  [B=-1]" : "");

	fprintf (helpout, "  C=0          same as --nochain --gapped%s\n",
	                 ((!defaultParams.chain) && (defaultParams.gappedExtend))? "             [C=0]" : "");
	fprintf (helpout, "  C=1          same as --chain   --nogapped%s\n",
	                 ((defaultParams.chain) && (!defaultParams.gappedExtend))? "           [C=1]" : "");
	fprintf (helpout, "  C=2          same as --chain   --gapped%s\n",
	                 ((defaultParams.chain) && (defaultParams.gappedExtend))? "             [C=2]" : "");
	fprintf (helpout, "  C=3          same as --nochain --nogapped%s\n",
	                 ((!defaultParams.chain) && (!defaultParams.gappedExtend))? "           [C=3]" : "");

	fprintf (helpout, "  c=1          same as --census%23s[c=%d]\n",
	                 "", (defaultParams.reportCensus)? 1 : 0);

	fprintf (helpout, "  E=<penalty>  same as --gap=<..,penalty>%13s[E=" scoreFmtSimple "]\n",
	                 "", HOXD70_extend);

	fprintf (helpout, "  G=<score>    same as --chain=<score,..>%13s[G=" scoreFmtSimple "]\n",
	                 "", defaultParams.chainDiag);

	fprintf (helpout, "  H=<score>    same as --inner=<score>%16s[H=" scoreFmtSimple "]\n",
	                 "", defaultParams.innerThreshold);

	fprintf (helpout, "  K=<score>    same as --hspthresh=<score>%12s[K=%s]\n",
	                 "", score_thresh_to_string (&defaultParams.hspThreshold));
	fprintf (helpout, "  L=<score>    same as --gappedthresh=<score>%9s[L=K]\n", "");

	fprintf (helpout, "  M=<count>    same as --masking=<count>%14s[M=%d]\n",
	                 "", defaultParams.dynamicMasking);

	fprintf (helpout, "  m=<bytes>    same as --traceback=<bytes>%12s[m=%s]\n",
	                 "", unitize(defaultParams.tracebackMem));

	fprintf (helpout, "  O=<penalty>  same as --gap=<penalty,..>%13s[O=" scoreFmtSimple "]\n",
	                 "", HOXD70_open);

	fprintf (helpout, "  P=0          same as --noentropy%s\n",
	                 (!defaultParams.entropicHsp)? "                    [P=0]" : "");
	fprintf (helpout, "  P=1          same as --entropy%s\n",
	                 ((defaultParams.entropicHsp) && (!defaultParams.reportEntropy))? "                      [P=1]" : "");
	fprintf (helpout, "  P>1          same as --entropy=report%s\n",
	                 ((defaultParams.entropicHsp) && (defaultParams.reportEntropy))? "               [P>1]" : "");

	fprintf (helpout, "  Q=<file>     same as --scores=<file>%16s[Q=<HOXD70>]\n","");

	fprintf (helpout, "  R=<score>    same as --chain=<..,score>%13s[R=" scoreFmtSimple "]\n",
	                 "", defaultParams.chainAnti);

	fprintf (helpout, "  T=1          same as --seed=12of19 --transition%s\n",
	                 ((defaultParams.withTrans == 1) && (strcmp(defaultSeedString,seed_12of19) == 0))? "     [T=1]" : "");
	fprintf (helpout, "  T=2          same as --seed=12of19 --notransition%s\n",
	                 ((defaultParams.withTrans == 0) && (strcmp(defaultSeedString,seed_12of19) == 0))? "   [T=2]" : "");
	fprintf (helpout, "  T=3          same as --seed=14of22 --transition%s\n",
	                 ((defaultParams.withTrans == 1) && (strcmp(defaultSeedString,seed_14of22) == 0))? "     [T=3]" : "");
	fprintf (helpout, "  T=4          same as --seed=14of22 --notransition%s\n",
	                 ((defaultParams.withTrans == 0) && (strcmp(defaultSeedString,seed_14of22) == 0))? "   [T=4]" : "");

	fprintf (helpout, "  U=1          same as --unitscores\n");

	fprintf (helpout, "  W=<length>   same as --seed=match(<length>)\n");

	fprintf (helpout, "  X=<score>    same as --xdrop=<score>%16s[X=10sub[A][A]]\n", "");
	fprintf (helpout, "  Y=<score>    same as --ydrop=<score>%16s[Y=O+300E]\n", "");

	fprintf (helpout, "  Z=<length>   same as --step=<length>%16s[Z=%u]\n",
	                 "", defaultParams.step);

	fprintf (helpout, "  v=0          same as --verbosity=0%s\n",
	                 (defaultParams.verbosity == 0)? "                  [v=0]" : "");
	fprintf (helpout, "  v=1          same as --verbosity=10%s\n",
	                 (defaultParams.verbosity == 10)? "                 [v=1]" : "");

	fprintf (helpout, "<HOXD70>\n");
	hoxScoring = new_dna_score_set (HOXD70, 0, 0, 0, 0);
	print_score_matrix_lf (helpout, hoxScoring, false, '\n');
	free_if_valid ("", hoxScoring);

	exit (1);
	}


static void chastise (const char* format, ...);
static void chastise (const char* format, ...)
	{
	va_list	args;

	va_start (args, format);
	if (format != NULL)
		vfprintf (stderr, format, args);
	va_end (args);

	usage ();
	}


// 'local' variables shared by parse_options() and parse_options_loop()

char*	seedString;
char*	scoreFileName;
char*	codeFileName;
char*	infControlFileName;
int		haveXDrop;
int		haveYDrop;
int		haveStep;
int		haveHspThreshold;
int		haveGappedThreshold;
int		haveInterpThreshold;
int		haveEntropicHsp;
int		haveBallScore;
int		haveWithTrans;
int		haveMaxIdentity;
int		useUnitScores;
int		gappedExtendVerbosity;
int		unitMatch;
int		unitMismatch;
int		haveGapOpen;
int		haveGapExtend;
char*	gapOpenStr;
char*	gapExtendStr;
score	gapOpen;
score	gapExtend;
int		twinsYes, minGap, maxGap;
double	ballScoreFactor;
int		haveSpecialSubs;
score	specialSubScores[4][4];
int		nIsAmbiguous;


// parse_options_loop-- main options parsing loop;  it is separate from the
//   overall options parsing routine because some options expand to a second
//   level of parsing

static void parse_options_loop   (int argc, char** argv,
                                  control* lzParams, control* izParams,
                                  int isToplevel);
static void parse_options_string (char* s,
                                  control* lzParams, control* izParams);

static void parse_options_loop
   (int			argc,
	char**		argv,
	control*	lzParams,
	control*	izParams,
	int			isTopLevel)
	{
	char*		argStr;
	int			argLen, argsLen;
	char*		scan, *scan2;
	int			items, scanned;
	int			wordLen, ix;
	float		minPctId, maxPctId;
	float		minCov,   maxCov;
	int			r, c;
	int			tempInt;

	argsLen = 0; // (placate compiler)

	while (argc > 0)
		{
		// copy argument (if it turns out to be a file name, we'll erase it
		// later)

		if (isTopLevel)
			{
			argsLen = strlen(lzParams->args);
			strcpy (lzParams->args+argsLen, argv[0]);
			strcpy (lzParams->args+argsLen+strlen(argv[0])," ");
			}

		// locate arg value string, if there is one

		argStr = strchr(argv[0],'=');
		if (argStr != NULL) argStr++;

		// --svn (unadvertised)
		// $$$ This needs to be improved so that it shows the *latest* revision
		// $$$ .. number of any module in the build path

		if (strcmp (argv[0], "--svn") == 0)
			{ printf ("SVN revision: %s\n", svnRevisionNumber);  exit (1); }

		// --self

		if (strcmp (argv[0], "--self") == 0)
			{ lzParams->selfCompare = true;  goto next_arg; }

		// --quantum

		if (strcmp (argv[0], "--quantum") == 0)
			{ lzParams->queryIsQuantum = true;  goto next_arg; }

		// --seed=<seed>, T=<code>, or W=<length>

		if (strcmp (argv[0], "T=0") == 0)	// in blastz, T=0 was accompanied
			{								// .. by W=<length>, which did not
			lzParams->withTrans = 0;			// .. support transitions
			goto next_arg;
			}

		if ((seedString == NULL) && (strcmp (argv[0], "T=1") == 0))
			{
			seedString = copy_string (seed_12of19);
			lzParams->withTrans = 1;
			goto next_arg;
			}

		if ((seedString == NULL) && (strcmp (argv[0], "T=2") == 0))
			{
			seedString = copy_string (seed_12of19);
			lzParams->withTrans = 0;
			goto next_arg;
			}

		if ((seedString == NULL) && (strcmp (argv[0], "T=3") == 0))
			{
			seedString = copy_string (seed_14of22);
			lzParams->withTrans = 1;
			goto next_arg;
			}

		if ((seedString == NULL) && (strcmp (argv[0], "T=4") == 0))
			{
			seedString = copy_string (seed_14of22);
			lzParams->withTrans = 0;
			goto next_arg;
			}

		if (strcmp_prefix (argv[0], "W=") == 0)
			{
			if (seedString != NULL)
				chastise ("can't specify W=<length> with --seed\n");
			items = sscanf (argStr, "%d", &wordLen);
			if (items != 1) goto cant_understand;
			lzParams->withTrans = 0;
			goto build_match_seed;
			}

		if ((seedString == NULL) && (strcmp_prefix (argv[0], "--seed=") == 0))
			{
			if (strcmp (argStr, "12of19") == 0)
				{
				seedString = copy_string (seed_12of19);
				goto next_arg;
				}

			if (strcmp (argStr, "14of22") == 0)
				{
				seedString = copy_string (seed_14of22);
				goto next_arg;
				}

			if ((strcmp_prefix (argStr, "match(") == 0)
			 && (argStr[strlen(argStr)-1] == ')'))
				{
				scan  = strchr(argStr,'(') + 1;
				items = sscanf (scan, "%d)", &wordLen);
				if (items != 1) goto cant_understand;
			build_match_seed:
				if ((wordLen < 2) || (wordLen > 15))
					chastise ("%d is not a valid word length\n", wordLen);

				seedString = malloc_or_die ("parse_options_loop (wordLen)", wordLen + 1);
				for (ix=0 ; ix<wordLen ; ix++)
					seedString[ix] = '1';
				seedString[wordLen] = 0;
				goto next_arg;
				}

			if ((strcmp_prefix (argStr, "half(") == 0)
			 && (argStr[strlen(argStr)-1] == ')'))
				{
				scan  = strchr(argStr,'(') + 1;
				items = sscanf (scan, "%d)", &wordLen);
				if (items != 1) goto cant_understand;
				if ((wordLen < 2) || (wordLen > 31))
					chastise ("%d is not a valid word length\n", wordLen);

				seedString = malloc_or_die ("parse_options_loop (wordLen)", wordLen + 1);
				for (ix=0 ; ix<wordLen ; ix++)
					seedString[ix] = 'T';
				seedString[wordLen] = 0;
				goto next_arg;
				}

			seedString = copy_string (argStr);
			goto next_arg;
			}

		// --[no]trans[ition]

		if ((strcmp (argv[0], "--notrans")      == 0)
		 || (strcmp (argv[0], "--notransition") == 0))
			{ lzParams->withTrans = 0;  goto next_arg; }

		if ((strcmp (argv[0], "--trans")        == 0)
		 || (strcmp (argv[0], "--transition")   == 0)
		 || (strcmp (argv[0], "--trans=1")      == 0)
		 || (strcmp (argv[0], "--transition=1") == 0))
			{ lzParams->withTrans = 1;  haveWithTrans = true;  goto next_arg; }

		if ((strcmp (argv[0], "--trans=2")      == 0)
		 || (strcmp (argv[0], "--transition=2") == 0))
			{ lzParams->withTrans = 2;  haveWithTrans = true;  goto next_arg; }

		// --[no]filter=<[T:]M>

		if (strcmp (argv[0], "--nofilter") == 0)
			{ lzParams->minMatches = -1;  goto next_arg; }

		if (strcmp_prefix (argv[0], "--filter=") == 0)
			{
			scan = strchr(argStr,':');
			if (scan == NULL)
				{
				lzParams->maxTransversions = 0;
				lzParams->minMatches       = string_to_int (argStr);
				}
			else
				{
				*(scan++) = 0;
				lzParams->maxTransversions = string_to_int (argStr);
				lzParams->minMatches       = string_to_int (scan);
				}

			goto next_arg;
			}

		// --word=<bits>

		if (strcmp_prefix (argv[0], "--word=") == 0)
			{
			lzParams->maxIndexBits = string_to_int (argStr);
			goto next_arg;
			}

		// --notwins and --twins=<min:max>

		if (strcmp (argv[0], "--notwins") == 0)
			{ twinsYes = false;  goto next_arg; }

		if (strcmp_prefix (argv[0], "--twins=") == 0)
			{
			twinsYes = true;
			scan = strchr(argStr,':');
			if (scan == NULL)
				{
				minGap = 0;
				maxGap = string_to_int (argStr);
				}
			else
				{
				*(scan++) = 0;
				minGap = string_to_int (argStr);
				maxGap = string_to_int (scan);
				}

			goto next_arg;
			}

		// --recoverhits

		if (strcmp (argv[0], "--recoverhits") == 0)
			{ lzParams->basicHitType = hitRecover;  goto next_arg; }

		// --rawhits

		if (strcmp (argv[0], "--rawhits") == 0)
			{ lzParams->noHitFiltering = true;  goto next_arg; }

		// --step=<length> or Z=<length>

		if ((strcmp_prefix (argv[0], "--step=") == 0)
		 || (strcmp_prefix (argv[0], "Z=")      == 0))
			{
			tempInt = string_to_int (argStr);
			if (tempInt <= 0)
				suicidef ("--step must be positive");
			lzParams->step = tempInt;
			haveStep       = true;
			goto next_arg;
			}

		// --both[strands], etc.

		if ((strcmp (argv[0], "--both")        == 0)
		 || (strcmp (argv[0], "--bothstrands") == 0))
			{ lzParams->whichStrand = 2;  goto next_arg; }

		if ((strcmp (argv[0], "--plus")       == 0)
		 || (strcmp (argv[0], "--plusstrand") == 0))
			{ lzParams->whichStrand = 0;  goto next_arg; }

		if ((strcmp (argv[0], "--minus")       == 0)
		 || (strcmp (argv[0], "--minusstrand") == 0))
			{ lzParams->whichStrand = -1;  goto next_arg; }

		if (strcmp_prefix (argv[0], "B=") == 0)
			{
			lzParams->whichStrand = string_to_int (argStr);
			goto next_arg;
			}

		// --backward[s] (unadvertised)

		if ((strcmp (argv[0], "--backward") == 0)
		 || (strcmp (argv[0], "--backwards") == 0))
			{ lzParams->backward = true;  goto next_arg; }

		// --ambiguousn

		if (strcmp (argv[0], "--ambiguousn") == 0)
			{ nIsAmbiguous = true;  goto next_arg; }

		// --[no]gfextend or (unadvertised) --[no]gfx

		if ((strcmp (argv[0], "--gfextend") == 0)
		 || (strcmp (argv[0], "--gfx"     ) == 0))
			{ lzParams->gfExtend = gfexXDrop;  goto next_arg; }

		if ((strcmp (argv[0], "--nogfextend") == 0)
		 || (strcmp (argv[0], "--nogfx"     ) == 0))
			{ lzParams->gfExtend = gfexNoExtend;  goto next_arg; }

		// --justhits or --hitsonly (unadvertised)

		if ((strcmp (argv[0], "--justhits") == 0)
		 || (strcmp (argv[0], "--hitsonly") == 0))
			{
			lzParams->gfExtend     = gfexNoExtend;
			lzParams->gappedExtend = false;
			goto next_arg;
			}

		// --[no]chain, --chain=<diag,anti>, G=, or R=

		if (strcmp (argv[0], "--chain") == 0)
			{ lzParams->chain = true;  goto next_arg; }

		if (strcmp (argv[0], "--nochain") == 0)
			{ lzParams->chain = false;  goto next_arg; }

		if (strcmp_prefix (argv[0], "--chain=") == 0)
			{
			lzParams->chain = true;
			scan = strchr(argStr,',');
			if (scan == NULL)
				chastise ("%s is not a valid pair of chain penalties\n", argStr);
			*scan = 0;
			lzParams->chainDiag = string_to_score (argStr);
			*(scan++) = ',';
			lzParams->chainAnti = string_to_score (scan);
			goto next_arg;
			}

		if (strcmp_prefix (argv[0], "G=") == 0)
			{
			lzParams->chainDiag = string_to_score (argStr);
			goto next_arg;
			}

		if (strcmp_prefix (argv[0], "R=") == 0)
			{
			lzParams->chainAnti = string_to_score (argStr);
			goto next_arg;
			}

		// --[no]gapped, C=<code>, or (unadvertised) --[no]gx

		if ((strcmp (argv[0], "--gapped") == 0)
		 || (strcmp (argv[0], "--gx"    ) == 0))
			{ lzParams->gappedExtend = true;  goto next_arg; }

		if ((strcmp (argv[0], "--nogapped") == 0)
		 || (strcmp (argv[0], "--ungapped") == 0)
		 || (strcmp (argv[0], "--nogx"    ) == 0))
			{ lzParams->gappedExtend = false;  goto next_arg; }

		if (strcmp (argv[0], "C=0") == 0)
			{ lzParams->chain = false; lzParams->gappedExtend = true;  goto next_arg; }

		if (strcmp (argv[0], "C=1") == 0)
			{ lzParams->chain = true;  lzParams->gappedExtend = false; goto next_arg; }

		if (strcmp (argv[0], "C=2") == 0)
			{ lzParams->chain = true;  lzParams->gappedExtend = true;  goto next_arg; }

		if (strcmp (argv[0], "C=3") == 0)
			{ lzParams->chain = false; lzParams->gappedExtend = false; goto next_arg; }

		// --score[s]=<file> or Q=<file>

		if ((scoreFileName == NULL)
		 && ((strcmp_prefix (argv[0], "--scores=") == 0)
		  || (strcmp_prefix (argv[0], "--score=")  == 0)
		  || (strcmp_prefix (argv[0], "Q=")        == 0)))
			{
			scoreFileName = copy_string (argStr);
			goto next_arg;
			}

		// --unitscore[s] or U=1
		// or (unadvertised) U=<match>[,<mismatch>]
		// or (unadvertised) --match=<match>[,<mismatch>]

		if ((strcmp (argv[0], "--unitscore")  == 0)
		 || (strcmp (argv[0], "--unitscores") == 0)
		 || (strcmp (argv[0], "U=1")          == 0))
			{
			useUnitScores = true;
			unitMatch     = 1;
			unitMismatch  = -1;
			goto next_arg;
			}

		if ((strcmp_prefix (argv[0], "--match=") == 0)
		 || (strcmp_prefix (argv[0], "U=")       == 0))
			{
			useUnitScores = true;
			scan = strchr(argStr,',');
			if (scan != NULL)
				*(scan++) = 0;
			unitMatch = string_to_score (argStr);
			if (unitMatch <= 0)
				chastise ("%s is not a valid match score\n", argStr);
			if (scan == NULL)
				unitMismatch = -unitMatch;
			else
				{
				unitMismatch = -string_to_score (scan);
				if (unitMismatch >= 0)
					chastise ("%s is not a valid mismatch penalty\n", scan);
				}
			goto next_arg;
			}

		// (unadvertised) <nuc1><nuc2>:<score>

		if ((nuc_to_bits[((u8*)argv[0])[0]] >= 0)
		 && (nuc_to_bits[((u8*)argv[0])[1]] >= 0)
		 && (argv[0][2] == ':'))
			{
			r = nuc_to_bits[(u8)argv[0][0]];
			c = nuc_to_bits[(u8)argv[0][1]];
			specialSubScores[r][c] = string_to_score (&argv[0][3]);
			haveSpecialSubs = true;
			goto next_arg;
			}

		// --infer[only][=<control>]

		if ((infControlFileName == NULL)
		 && (strcmp (argv[0], "--infer") == 0))
			{
			lzParams->inferScores = true;
			lzParams->inferOnly   = false;
			goto next_arg;
			}

		if ((infControlFileName == NULL)
		 && (strcmp_prefix (argv[0], "--infer=") == 0))
			{
			lzParams->inferScores = true;
			lzParams->inferOnly   = false;
			infControlFileName = copy_string (argStr);
			goto next_arg;
			}

		if ((infControlFileName == NULL)
		 && (strcmp (argv[0], "--inferonly") == 0))
			{
			lzParams->inferScores = true;
			lzParams->inferOnly   = true;
			goto next_arg;
			}

		if ((infControlFileName == NULL)
		 && (strcmp_prefix (argv[0], "--inferonly=") == 0))
			{
			lzParams->inferScores = true;
			lzParams->inferOnly   = true;
			infControlFileName = copy_string (argStr);
			goto next_arg;
			}

		// --infscores[=<file>]

		if (strcmp (argv[0], "--infscores") == 0)
			{
			lzParams->inferScores = true;
			goto next_arg;
			}

		if ((izParams->ic.inferFileName == NULL)
		 && (strcmp_prefix (argv[0], "--infscores=") == 0))
			{
			lzParams->inferScores      = true;
			izParams->ic.inferFileName = copy_string (argStr);
			goto next_arg;
			}

		// --gap=<[open,]extend> or O=<open> or E=<extend>

		if (strcmp_prefix (argv[0], "--gap=") == 0)
			{
			scan = strchr(argStr,',');
			if (scan == NULL)
				{
				gapExtend = string_to_score (argStr);
				if (gapExtend < 0)
					chastise ("%s is not a valid gap extension penalty\n", argStr);
				haveGapExtend = true;
				gapExtendStr  = argStr;
				}
			else
				{
				*(scan++) = 0;
				gapOpen      = string_to_score (argStr);
				gapExtend    = string_to_score (scan);
				haveGapOpen  = haveGapExtend = true;
				gapOpenStr   = argStr;
				gapExtendStr = scan;
				}
			goto next_arg;
			}

		if (strcmp_prefix (argv[0], "O=") == 0)
			{
			gapOpen = string_to_score (argStr);
			haveGapOpen = true;
			gapOpenStr  = argStr;
			goto next_arg;
			}

		if (strcmp_prefix (argv[0], "E=") == 0)
			{
			gapExtend = string_to_score (argStr);
			haveGapExtend = true;
			gapExtendStr  = argStr;
			goto next_arg;
			}

		// --xdrop=<score> or X=<score>

		if ((strcmp_prefix (argv[0], "--xdrop=") == 0)
		 || (strcmp_prefix (argv[0], "X=")       == 0))
			{
			lzParams->xDrop = string_to_score (argStr);
			haveXDrop = true;
			goto next_arg;
			}

		// --ydrop=<score> or Y=<score>

		if ((strcmp_prefix (argv[0], "--ydrop=") == 0)
		 || (strcmp_prefix (argv[0], "Y=")       == 0))
			{
			lzParams->yDrop = string_to_score (argStr);
			haveYDrop = true;
			goto next_arg;
			}

		// --hspthresh=<score_threshold> or K=<score_threshold>

		if ((strcmp_prefix (argv[0], "--hspthresh=") == 0)
		 || (strcmp_prefix (argv[0], "--mspthresh=") == 0)
		 || (strcmp_prefix (argv[0], "K="          ) == 0))
			{
			lzParams->hspThreshold = string_to_score_thresh (argStr);
			haveHspThreshold = true;
			goto next_arg;
			}

		// --exact=<length>

		if (strcmp_prefix (argv[0], "--exact=") == 0)
			{
			lzParams->gfExtend = gfexExact;
			lzParams->hspThreshold.t = 'S';
			lzParams->hspThreshold.s = string_to_score (argStr);
			if (lzParams->hspThreshold.s <= 0)
				chastise ("%s is not a valid exact match threshold\n", argStr);
			haveHspThreshold = true;
			goto next_arg;
			}

		// --inner=<score> or H=<score>

		if ((strcmp_prefix (argv[0], "--inner=") == 0)
		 || (strcmp_prefix (argv[0], "H="      ) == 0))
			{
			lzParams->innerThreshold = string_to_score (argStr);
			haveInterpThreshold = true;
			goto next_arg;
			}

		// --gappedthresh=<score_threshold> or L=<score_threshold>

		if ((strcmp_prefix (argv[0], "--gappedthresh=") == 0)
		 || (strcmp_prefix (argv[0], "L="             ) == 0))
			{
			lzParams->gappedThreshold = string_to_score_thresh (argStr);
			haveGappedThreshold = true;
			goto next_arg;
			}

		// --ball=<score>

		if (strcmp_prefix (argv[0], "--ball=") == 0)
			{
			argLen = strlen(argStr);
			if ((argLen > 0) && (argStr[argLen-1] == '%'))
				{
				lzParams->ballScore = 0;		// (just signals that --ball used)
				ballScoreFactor = pct_string_to_double (argStr);
				}
			else
				{
				lzParams->ballScore = string_to_score (argStr);
				haveBallScore = true;
				}
			goto next_arg;
			}

		// --[no]entropy or P=<flag>

		if ((strcmp (argv[0], "--entropy") == 0)
		 || (strcmp (argv[0], "P=1"      ) == 0))
			{ lzParams->entropicHsp = haveEntropicHsp = true;  goto next_arg; }

		if ((strcmp (argv[0], "--noentropy") == 0)
		 || (strcmp (argv[0], "P=0"        ) == 0))
			{ lzParams->entropicHsp = haveEntropicHsp = false;  goto next_arg; }

		if (strcmp_prefix (argv[0], "P=") == 0)
			{
			if (string_to_int (argStr) <= 0)
				chastise ("illegal value for P");
			goto report_entropy;
			}

		if (strcmp (argv[0], "--entropy=report") == 0)
			{
		report_entropy:
			lzParams->entropicHsp = lzParams->reportEntropy = haveEntropicHsp = true;
			goto next_arg;
			}

		// --traceback=<bytes> or m=<bytes>

		if ((strcmp_prefix (argv[0], "--traceback=") == 0)
		 || (strcmp_prefix (argv[0], "m="          ) == 0))
			{
			lzParams->tracebackMem = string_to_int2 (argStr, false);
			goto next_arg;
			}

		// --masking=<count> or M=<count>

		if ((strcmp_prefix (argv[0], "--masking=") == 0)
		 || (strcmp_prefix (argv[0], "M="        ) == 0))
			{
			tempInt = string_to_int (argStr);
			if (tempInt < 0)
				suicidef ("--masking cannot be negative");
			else if (tempInt >= 255)  // (255 itself is not allowed)
				suicidef ("--masking must be less than 255");
			lzParams->dynamicMasking = tempInt;
			goto next_arg;
			}

		// --[no]census or c=<flag>

		if (strcmp_prefix (argv[0], "c=") == 0)
			{
			if (string_to_int (argStr) == 0) goto census_off;
			                            else goto census_on;
			}

		if (strcmp (argv[0], "--census") == 0)
			{
		census_on:
			lzParams->reportCensus = true;
			goto next_arg;
			}

		if (strcmp (argv[0], "--nocensus") == 0)
			{
		census_off:
			lzParams->reportCensus = false;
			goto next_arg;
			}

		// --identity=min[..max]

		if (strcmp_prefix (argv[0], "--identity=") == 0)
			{
			if (strcmp (argStr,"..") == 0) goto cant_understand;
			scan  = argStr;
			scan2 = strstr (scan, "..");
			if      (scan2 == NULL) { ;                           } // min
			else if (scan2 == scan) { scan2 += 2;    scan = NULL; } // ..max
			else if (scan2[2] == 0) { scan2 = NULL;               } // min..
			else                    { scan2 += 2;                 } // min..max

			minPctId = 0.0;
			if (scan != NULL)
				{
				scanned = -1;
				sscanf (scan, "%f%n", &minPctId, &scanned);
				if (scanned == -1) goto cant_understand;
				scan += scanned;
				if (*scan == '%') scan++;
				if (scan2 == NULL)
					{
					if ((*scan != 0) && (strcmp (scan, ".") != 0) && (strcmp (scan, "..") != 0))
						goto cant_understand;
					}
				else
					{
					if (*scan != '.') goto cant_understand;  scan++;
					if (scan != scan2) { if (*scan != '.') goto cant_understand;  scan++; }
					if (scan != scan2) goto cant_understand;
					}
				}

			maxPctId = 100.0;
			if (scan2 != NULL)
				{
				scanned = -1;
				sscanf (scan2, "%f%n", &maxPctId, &scanned);
				if (scanned == -1) goto cant_understand;
				scan2 += scanned;
				if (*scan2 == '%') scan2++;
				if (*scan2 != 0) goto cant_understand;
				}

			if ((minPctId < 0) || (maxPctId > 100) || (minPctId > maxPctId))
				goto cant_understand;

			lzParams->minIdentity = minPctId / 100.0;
			lzParams->maxIdentity = maxPctId / 100.0;
			haveMaxIdentity       = true;
			goto next_arg;
			}

		// --coverage=min[..max]

		if (strcmp_prefix (argv[0], "--coverage=") == 0)
			{
			if (strcmp (argStr,"..") == 0) goto cant_understand;
			scan  = argStr;
			scan2 = strstr (scan, "..");
			if      (scan2 == NULL) { ;                           } // min
			else if (scan2 == scan) { scan2 += 2;    scan = NULL; } // ..max
			else if (scan2[2] == 0) { scan2 = NULL;               } // min..
			else                    { scan2 += 2;                 } // min..max

			minCov = 0.0;
			if (scan != NULL)
				{
				scanned = -1;
				sscanf (scan, "%f%n", &minCov, &scanned);
				if (scanned == -1) goto cant_understand;
				scan += scanned;
				if (*scan == '%') scan++;
				if (scan2 == NULL)
					{
					if ((*scan != 0) && (strcmp (scan, ".") != 0) && (strcmp (scan, "..") != 0))
						goto cant_understand;
					}
				else
					{
					if (*scan != '.') goto cant_understand;  scan++;
					if (scan != scan2) { if (*scan != '.') goto cant_understand;  scan++; }
					if (scan != scan2) goto cant_understand;
					}
				}

			maxCov = 100.0;
			if (scan2 != NULL)
				{
				scanned = -1;
				sscanf (scan2, "%f%n", &maxCov, &scanned);
				if (scanned == -1) goto cant_understand;
				scan2 += scanned;
				if (*scan2 == '%') scan2++;
				if (*scan2 != 0) goto cant_understand;
				}

			if ((minCov < 0) || (maxCov > 100) || (minCov > maxCov))
				goto cant_understand;

			lzParams->minCoverage = minCov / 100.0;
			lzParams->maxCoverage = maxCov / 100.0;
			goto next_arg;
			}

		// --[no]mirror

		if (strcmp (argv[0], "--mirror") == 0)
			{ lzParams->mirror = true;  goto next_arg; }

		if (strcmp (argv[0], "--nomirror") == 0)
			{ lzParams->mirror = false;  goto next_arg; }

		// --code=<file>

		if ((codeFileName == NULL)
		 && (strcmp_prefix (argv[0], "--code=") == 0))
			{
			codeFileName = copy_string (argStr);
			goto next_arg;
			}

		// --format=<format> (many variations)

		if ((strcmp (argv[0], "--format=gfa") == 0)
		 || (strcmp (argv[0], "--format=GFA") == 0)
		 || (strcmp (argv[0], "--gfa")        == 0)
		 || (strcmp (argv[0], "--GFA")        == 0))
			{ lzParams->outputFormat = fmtGfa;  goto next_arg; }

		if ((strcmp (argv[0], "--format=gfanoscore") == 0) // (unadvertised)
		 || (strcmp (argv[0], "--format=GFANOSCORE") == 0)
		 || (strcmp (argv[0], "--gfanoscore")        == 0)
		 || (strcmp (argv[0], "--GFANOSCORE")        == 0))
			{ lzParams->outputFormat = fmtGfaNoScore;  goto next_arg; }

		if ((strcmp (argv[0], "--format=lav") == 0)
		 || (strcmp (argv[0], "--format=LAV") == 0)
		 || (strcmp (argv[0], "--lav")        == 0)
		 || (strcmp (argv[0], "--LAV")        == 0))
			{ lzParams->outputFormat = fmtLav;  goto next_arg; }

		if ((strcmp (argv[0], "--format=lav+") == 0)
		 || (strcmp (argv[0], "--format=LAV+") == 0)
		 || (strcmp (argv[0], "--lav+")        == 0)
		 || (strcmp (argv[0], "--LAV+")        == 0))
			{ lzParams->outputFormat = fmtLavComment;  goto next_arg; }

		if ((strcmp (argv[0], "--format=lav+text") == 0)
		 || (strcmp (argv[0], "--format=LAV+text") == 0)
		 || (strcmp (argv[0], "--lav+text")        == 0)
		 || (strcmp (argv[0], "--LAV+text")        == 0)
		 || (strcmp (argv[0], "--format=text+lav") == 0)
		 || (strcmp (argv[0], "--format=text+LAV") == 0)
		 || (strcmp (argv[0], "--text+lav")        == 0)
		 || (strcmp (argv[0], "--text+LAV")        == 0))
			{ lzParams->outputFormat = fmtLavText;  goto next_arg; }

		if ((strcmp (argv[0], "--format=lavscore") == 0) // (unadvertised)
		 || (strcmp (argv[0], "--format=LAVSCORE") == 0)
		 || (strcmp (argv[0], "--lavscore")        == 0)
		 || (strcmp (argv[0], "--LAVSCORE")        == 0))
			{ lzParams->outputFormat = fmtLavScore;  goto next_arg; }

		if ((strcmp (argv[0], "--format=axt") == 0)
		 || (strcmp (argv[0], "--format=AXT") == 0)
		 || (strcmp (argv[0], "--axt")        == 0)
		 || (strcmp (argv[0], "--AXT")        == 0))
			{ lzParams->outputFormat = fmtAxt;  goto next_arg; }

		if ((strcmp (argv[0], "--format=axt+") == 0)
		 || (strcmp (argv[0], "--format=AXT+") == 0)
		 || (strcmp (argv[0], "--axt+")        == 0)
		 || (strcmp (argv[0], "--AXT+")        == 0))
			{ lzParams->outputFormat = fmtAxtComment;  goto next_arg; }

		if ((strcmp (argv[0], "--format=maf") == 0)
		 || (strcmp (argv[0], "--format=MAF") == 0)
		 || (strcmp (argv[0], "--maf")        == 0)
		 || (strcmp (argv[0], "--MAF")        == 0))
			{
			lzParams->outputFormat = fmtMaf;
			maf_distinguishNames   = false;
			goto next_arg;
			}

		if ((strcmp (argv[0], "--format=~maf") == 0)
		 || (strcmp (argv[0], "--format=~MAF") == 0))
			{
			lzParams->outputFormat = fmtMaf;
			maf_distinguishNames   = true;
			goto next_arg;
			}

		if ((strcmp (argv[0], "--format=maf+") == 0)
		 || (strcmp (argv[0], "--format=MAF+") == 0)
		 || (strcmp (argv[0], "--maf+")        == 0)
		 || (strcmp (argv[0], "--MAF+")        == 0))
			{
			lzParams->outputFormat = fmtMafComment;
			maf_distinguishNames   = false;
			goto next_arg;
			}

		if ((strcmp (argv[0], "--format=~maf+") == 0)
		 || (strcmp (argv[0], "--format=~MAF+") == 0))
			{
			lzParams->outputFormat = fmtMafComment;
			maf_distinguishNames   = true;
			goto next_arg;
			}

		if ((strcmp (argv[0], "--format=maf-") == 0)
		 || (strcmp (argv[0], "--format=MAF-") == 0)
		 || (strcmp (argv[0], "--maf-")        == 0)
		 || (strcmp (argv[0], "--MAF-")        == 0))
			{
			lzParams->outputFormat = fmtMafNoComment;
			maf_distinguishNames   = false;
			goto next_arg;
			}

		if (strcmp (argv[0], "--format=text") == 0)
			{ lzParams->outputFormat = fmtText;  goto next_arg; }

		if (strcmp (argv[0], "--format=comp") == 0)
			{ lzParams->outputFormat = fmtHspComp;  goto next_arg; }

		if ((strcmp (argv[0], "--format=istats")   == 0)
		 || (strcmp (argv[0], "--format=infstats") == 0))
			{
			if (haveMaxIdentity) goto set_inf_stats_format;
			maxPctId = 70.0;
			goto set_inf_stats_info;
			}

		if (((strcmp_prefix (argv[0], "--format=istats(")   == 0)
		  || (strcmp_prefix (argv[0], "--format=infstats(") == 0))
		 && (argv[0][strlen(argv[0])-1] == ')'))
			{
			scan = strchr(argv[0],'(') + 1;
			sscanf (scan, "%f%n", &maxPctId, &scanned);
			if (scanned == -1) goto cant_understand;
			scan += scanned;
			if (*scan == '%') scan++;
			if (strcmp (scan, ")") != 0) goto cant_understand;
			if ((maxPctId < 0) || (maxPctId > 100)) goto cant_understand;
		set_inf_stats_info:
			lzParams->maxIdentity  = maxPctId / 100.0;
			haveMaxIdentity        = true;
		set_inf_stats_format:
			lzParams->outputFormat = fmtInfStats;
			goto next_arg;
			}

		if (strcmp (argv[0], "--format=identity") == 0)
			{ lzParams->outputFormat = fmtIdDist;  goto next_arg; }

		if (strcmp (argv[0], "--format=deseed") == 0)
			{ lzParams->outputFormat = fmtDeseed;  goto next_arg; }

		if (strcmp (argv[0], "--format=none") == 0)
			{ lzParams->outputFormat = fmtNone;  goto next_arg; }

		// --[no]laj

		if (strcmp (argv[0], "--laj") == 0)
			{ lzParams->lajCompatible = true;  goto next_arg; }

		if (strcmp (argv[0], "--nolaj") == 0)
			{ lzParams->lajCompatible = false;  goto next_arg; }

		// (unadvertised) --expand=<bp>

		if (strcmp_prefix (argv[0], "--expand=") == 0)
			{
			tempInt = string_to_int (argStr);
			if (tempInt < 0)
				suicidef ("--expand cannot be negative");
			else if (tempInt >= 1000)
				suicidef ("--expand must be less than 1000");
			lzParams->textContext = (u32) tempInt;
			goto next_arg;
			}

		// precanned expansion arguments

		for (ix=0 ; ix<numExpanders ; ix++)
			{
			if (strcmp (argv[0], expanders[ix].argName) != 0) continue;
			parse_options_string (expanders[ix].expansion, lzParams, izParams);
			goto next_arg;
			}

		// --verbosity=<level>

		if (strcmp_prefix (argv[0], "v=0") == 0)
			{ lzParams->verbosity = 0;  goto next_arg; }

		if (strcmp_prefix (argv[0], "v=1") == 0)
			{ lzParams->verbosity = 10;  goto next_arg; }

		if (strcmp_prefix (argv[0], "--verbosity=") == 0)
			{
			lzParams->verbosity = string_to_int (argStr);
			if      (lzParams->verbosity < 0 ) lzParams->verbosity = 0;
			else if (lzParams->verbosity > 10) lzParams->verbosity = 10;
			goto next_arg;
			}

		if (strcmp_prefix (argv[0], "--gexverbosity=") == 0) // (unadvertised)
			{
			gappedExtendVerbosity = string_to_int (argStr);
			if      (gappedExtendVerbosity < 0 ) gappedExtendVerbosity = 0;
			else if (gappedExtendVerbosity > 10) gappedExtendVerbosity = 10;
			goto next_arg;
			}

		// --[no]runtime

		if (strcmp (argv[0], "--runtime") == 0)
			{ lzParams->reportTiming = true;  goto next_arg; }

		if (strcmp (argv[0], "--noruntime") == 0)
			{ lzParams->reportTiming = false;  goto next_arg; }

		// --tableonly[=count] and (unadvertised) --showtable[=count]

		if (strcmp (argv[0], "--tableonly") == 0)
			{
			lzParams->doSeedSearch = false;
			lzParams->showPosTable = 1;
			goto next_arg;
			}

		if (strcmp (argv[0], "--tableonly=count") == 0)
			{
			lzParams->doSeedSearch = false;
			lzParams->showPosTable = 2;
			goto next_arg;
			}

		if (strcmp (argv[0], "--showtable") == 0)
			{ lzParams->showPosTable = 1;  goto next_arg; }

		if (strcmp (argv[0], "--showtable=count") == 0)
			{ lzParams->showPosTable = 2;  goto next_arg; }

		// --[no]stats[=<file>] or (unadvertised) --stats=

		if (strcmp (argv[0], "--stats=") == 0)
			{ lzParams->reportStats = true;  goto next_arg; }

		if (strcmp (argv[0], "--stats") == 0)
			{ lzParams->showStats = true;  goto next_arg; }

		if (strcmp (argv[0], "--nostats") == 0)
			{ lzParams->showStats = false;  goto next_arg; }

		if ((lzParams->statsFileName == NULL)
		 && (strcmp_prefix (argv[0], "--stats=") == 0))
			{
			lzParams->statsFileName = copy_string (argStr);
			lzParams->showStats     = true;
			goto next_arg;
			}

		// --notruncationreport (unadvertised)

		if (strcmp (argv[0], "--notruncationreport") == 0)
			{ gapped_extend_inhibitTruncationReport = true;  goto next_arg; }

		// --version

		if (strcmp (argv[0], "--version") == 0)
			{
			fprintf (helpout, "%s (version %s.%s.%s built %s)",
			                  programName,
			                  programVersionMajor, programVersionMinor, programVersionSubMinor, programRevisionDate);

			#if (scoreType == 'I')
			fprintf (helpout, " score=int");
			#elif (scoreType == 'F')
			fprintf (helpout, " score=float");
			#elif (scoreType == 'D')
			fprintf (helpout, " score=double-float");
			#endif

			fprintf (helpout, ", sequence=%d-bit", maxSequenceIndex);

			fprintf (helpout, "\n");
			exit (1);
			}

		// --help

		if ((strcmp (argv[0], "--all")  == 0)
		 || (strcmp (argv[0], "--help") == 0)
		 || (strcmp (argv[0], "--help=all") == 0))
			{ all_options(); }

		// --help=files

		if (strcmp (argv[0], "--help=files") == 0)
			{ file_options(); }

		// --help=short[cuts]

		if ((strcmp (argv[0], "--short") == 0)
		 || (strcmp (argv[0], "--shortcuts") == 0)
		 || (strcmp (argv[0], "--help=short") == 0)
		 || (strcmp (argv[0], "--help=shortcuts") == 0)
		 || (strcmp (argv[0], "--blastz") == 0)
		 || (strcmp (argv[0], "--help=blastz") == 0))
			{ shortcuts(); }

		// --debug=<level> (unadvertised)

		if (strcmp (argv[0], "--debug") == 0)
			{ debug = 100;  goto next_arg; }

		if (strcmp (argv[0], "--debug=rawhits") == 0)
			{ seed_search_dbgShowRawHits = true;  goto next_arg; }

		if (strcmp (argv[0], "--debug=words") == 0)
			{ pos_table_dbgShowWords = true;  goto next_arg; }

		if (strcmp (argv[0], "--debug=seedhits") == 0)
			{ seed_search_dbgShowHits = true;  goto next_arg; }

		if (strcmp (argv[0], "--debug=currParams") == 0)
			{ dbgShowParams = true;  goto next_arg; }

		if (strcmp (argv[0], "--debug=hsps") == 0)
			{ dbgShowHsps = true;  goto next_arg; }

		if (strcmp (argv[0], "--debug=anchors") == 0)
			{ dbgShowAnchors = true;  goto next_arg; }

		if (strcmp (argv[0], "--debug=reduction") == 0)
			{ dbgInhibitSegmentReduction = true;  goto next_arg; }

		if (strcmp (argv[0], "--debug=pctid") == 0)
			{
			gapped_extend_dbgShowIdentity = true;
			identity_dist_dbgShowIdentity = true;
			infer_scores_dbgShowIdentity  = true;
			goto next_arg;
			}

		if (strcmp_prefix (argv[0], "--debug=queryprogress=") == 0)
			{
			dbgQueryProgress = string_to_int (strchr(argStr,'=')+1);
			goto next_arg;
			}

		if ((strcmp (argv[0], "--debug=converge") == 0)
		 || (strcmp (argv[0], "--debug=convergence") == 0))
			{ infer_scores_watchConverge = true;  goto next_arg; }

		if (strcmp (argv[0], "--debug=lav+infer") == 0)
			{ infer_scores_outputLav = true;  goto next_arg; }

		if (strcmp_prefix (argv[0], "--debug=") == 0)
			{
			debug = string_to_int (argStr);
			if      (debug < 0  ) debug = 0;
			else if (debug > 100) debug = 100;
			goto next_arg;
			}

		// unknown -- argument

		if (strcmp_prefix (argv[0], "--") == 0)
			goto cant_understand;

		// target file name

		if (lzParams->seq1Filename == NULL)
			{
			lzParams->seq1Filename = copy_string (argv[0]);
			goto next_arg_no_copy;
			}

		// query file name

		if (lzParams->seq2Filename == NULL)
			{
			lzParams->seq2Filename = copy_string (argv[0]);
			goto next_arg_no_copy;
			}

	cant_understand:
		chastise ("Can't understand \"%s\"\n", argv[0]);

	next_arg_no_copy:
		if (isTopLevel) lzParams->args[argsLen] = 0; // 'erase' the argument
	next_arg:
		argv++;  argc--;
		continue;

	//next_arg2:
	//	argv+=2;  argc-=2;
	//	continue;
		}
	}

// parse_options_string-- parse options from a string

static void parse_options_string
   (char*		s,
	control*	lzParams,
	control*	izParams)
	{
	int			argC, ix;
	char**		argV, *argS, *scan, *argEnd;

	for (scan=skip_whitespace(s),argC=0 ; *scan!=0 ; argC++)
		{
		scan = skip_darkspace  (scan);
		scan = skip_whitespace (scan);
		}

	argV = malloc_or_die ("parse_options_string (argV)", argC * sizeof(char*));
	argS = copy_string (s);

	for (scan=skip_whitespace(argS),ix=0 ; *scan!=0 ; ix++)
		{
		argV[ix] = scan;
		scan = skip_darkspace  (scan);  argEnd = scan;
		scan = skip_whitespace (scan);
		*argEnd = 0;
		}

	parse_options_loop (argC, argV, lzParams, izParams, /*top level*/ false);

	free_if_valid ("parse_options_string (argV)", argV);
	free_if_valid ("parse_options_string (argS)", argS);
	}

// parse_options-- overall options parsing

static void parse_options
   (int			_argc,
	char**		_argv,
	control*	lzParams,
	control*	izParams)
	{
	int			argc;
	char**		argv;
	int			argsLen, ix;
	score		maxScore;
	int			r, c;
	u8			nuc1, nuc2;
	char*		seq2Filename;
	score		myUnitScores[4][4];
	exscoreset*	xss;

	// skip program name

	argv = _argv+1;  argc = _argc - 1;

	//////////
	// set defaults
	//////////

	*izParams = defaultParams;
	*lzParams = defaultParams;

	seedString          = NULL;
	scoreFileName       = NULL;
	codeFileName        = NULL;
	infControlFileName  = NULL;
	haveXDrop           = false;
	haveYDrop           = false;
	haveStep            = false;
	haveHspThreshold    = false;
	haveGappedThreshold = false;
	haveInterpThreshold = false;
	haveEntropicHsp     = false;
	haveBallScore       = false;
	haveWithTrans       = false;
	haveMaxIdentity     = false;
	useUnitScores       = false;
	gappedExtendVerbosity = -1;
	unitMatch           = 1;
	unitMismatch        = -1;
	haveGapOpen         = false;
	haveGapExtend       = false;
	gapOpenStr          = NULL;
	gapExtendStr        = NULL;
	gapOpen             = 0;
	gapExtend           = 0;
	twinsYes            = defaultTwinsYes;
	minGap              = defaultTwinMinGap;
	maxGap              = defaultTwinMaxGap;
	ballScoreFactor     = defaultBallScoreFactor;
	haveSpecialSubs     = false;
	nIsAmbiguous        = false;

	// create a string to copy arguments into (this will be slightly bigger
	// than we need)

	argsLen = 2;
	for (ix=0 ; ix<argc ; ix++)
		argsLen += 1 + strlen(argv[ix]);

	lzParams->args = malloc_or_die ("parse_options (lzParams->args)", argsLen);
	lzParams->args[0] = 0;

	for (r=0 ; r<4 ; r++)
			for (c=0 ; c<4 ; c++)
		specialSubScores[r][c] = worstPossibleScore;

	//////////
	// scan arguments
	//////////

	parse_options_loop (argc, argv, lzParams, izParams, /*top level*/ true);

	//////////
	// check for sensibility
	//////////

#ifdef disallowEntropy
	if (lzParams->entropicHsp)
		chastise ("--entropy is currently disabled\n");
#endif

#ifndef collect_stats
	if (lzParams->showStats)
		chastise ("--stats is not implemented in this build of the program\n");
#endif // collect_stats

	if (lzParams->seq1Filename == NULL)
		chastise ("You must specify a target file\n");

	if (lzParams->inferOnly)
		{
		if (lzParams->backward)
			chastise ("--backward can't be used with --inferonly\n");

		if (lzParams->noHitFiltering)
			chastise ("--rawhits can't be used with --inferonly\n");

		if (lzParams->dynamicMasking > 0)
			chastise ("--masking can't be used with --inferonly\n");

		if (lzParams->reportCensus)
			chastise ("--census can't be used with --inferonly\n");

		if (lzParams->outputFormat != defaultParams.outputFormat)
			chastise ("--format=%s can't be used with --inferonly\n",
			          formatNames[lzParams->outputFormat]);

		if (lzParams->innerThreshold > 0)
			chastise ("--inner can't be used with --inferonly\n");
		}

	if (lzParams->selfCompare)
		{
		if (lzParams->seq2Filename != NULL)
			chastise ("--self can't be used when you specify a query file\n");
		lzParams->seq2Filename = copy_string (lzParams->seq1Filename);
		if (lzParams->mirror == -1)
			lzParams->mirror = true;
		}
	else if (lzParams->mirror == true)
		chastise ("--mirror can only be used with --self\n");
	else
		lzParams->mirror = false;

	if (!lzParams->doSeedSearch)
		{
		if (lzParams->seq2Filename != NULL)
			chastise ("--tableonly can't be used when you specify a query file\n");
		if (lzParams->inferScores)
			chastise ("--infer and --tableonly are not compatible\n");
		}

	if (lzParams->maxIndexBits < 8)
		chastise ("--word doesn't allow so few bits (%d)\n",
				  lzParams->maxIndexBits);

	if (lzParams->tracebackMem < 100*1024)
		chastise ("--traceback must be at least 100K (it's only %s)\n",
				  unitize(lzParams->tracebackMem));

	//////////
	// set up score set
	//////////

	if (lzParams->inferScores)
		{
		if (scoreFileName != NULL)
			chastise ("can't use --infer[only] and --scores together\n");

		if (useUnitScores)
			chastise ("can't use --infer[only] and --unitscores together\n");

		if (haveSpecialSubs)
			chastise ("can't use --infer[only] and special substitution scores together\n");
		}

	if (lzParams->gfExtend == gfexNoExtend)
		{
		if ((!lzParams->gappedExtend) && (scoreFileName != NULL))
			chastise ("--scores requires --gfextend or --gapped\n");

		if (haveXDrop)
			chastise ("--xdrop requires --gfextend\n");

		if (haveHspThreshold)
			chastise ("--hspthresh requires --gfextend\n");

		if (haveEntropicHsp)
			chastise ("--entropy requires --gfextend\n");

		lzParams->xDrop          = 0;
		lzParams->hspThreshold.t = 'S';
		lzParams->hspThreshold.s = 0;
		lzParams->entropicHsp    = false;
		}

	if (!lzParams->chain)
		{
		if (lzParams->chainDiag != 0)
			chastise ("G=<score> requires --chain\n");

		if (lzParams->chainAnti != 0)
			chastise ("R=<score> requires --chain\n");
		}

	if (!lzParams->gappedExtend)
		{
		if ((haveGapOpen) || (haveGapExtend))
			chastise ("--gap (or O= or E=) requires --gapped\n");

		if (haveYDrop)
			chastise ("--ydrop requires --gapped\n");

		if (haveGappedThreshold)
			chastise ("--gappedThreshold requires --gapped\n");

		if (haveInterpThreshold)
			chastise ("--inner requires --gapped\n");
		}

	if (haveXDrop && (lzParams->xDrop <= 0))
		chastise ("%d is not a legitimate x-drop threshold\n", lzParams->xDrop);

	if (haveYDrop && (lzParams->yDrop <= 0))
		chastise ("%d is not a legitimate y-drop threshold\n", lzParams->yDrop);

	if ((useUnitScores) && (scoreFileName != NULL))
		chastise ("can't use --unitscores and --scores together\n");

	if (scoreFileName != NULL)
		{
		// read scores and score-related paramters from file, allowing them
		// to be overridden by the command line

		xss = read_score_set_by_name (scoreFileName);
		lzParams->scoring = (scoreset*) xss;

		if (xss->seedSet)
			{
			if (seedString == NULL) // it contains params in command-line syntax
				parse_options_loop (1, &xss->seed, lzParams, izParams, /*top level*/ false);
			free_if_valid ("xss->seed", xss->seed);  xss->seed = NULL;
			}

		if (!haveGapOpen)
			gapOpen = lzParams->scoring->gapOpen;
		else
			{
			lzParams->scoring->gapOpen    = gapOpen;
			lzParams->scoring->gapOpenSet = true;
			}

		if (!haveGapExtend)
			gapExtend = lzParams->scoring->gapExtend;
		else
			{
			lzParams->scoring->gapExtend    = gapExtend;
			lzParams->scoring->gapExtendSet = true;
			}

		if ((!haveHspThreshold) && (xss->hspThresholdSet))
			{
			lzParams->hspThreshold.t    = 'S';
			lzParams->hspThreshold.s    = xss->hspThreshold;
			haveHspThreshold            = true;
			}
		if ((!haveGappedThreshold) && (xss->gappedThresholdSet))
			{
			lzParams->gappedThreshold.t = 'S';
			lzParams->gappedThreshold.s = xss->gappedThreshold;
			haveGappedThreshold         = true;
			}
		if ((!haveXDrop) && (xss->xDropSet))
			{
			lzParams->xDrop             = xss->xDrop;
			haveXDrop                   = true;
			}
		if ((!haveYDrop) && (xss->yDropSet))
			{
			lzParams->yDrop             = xss->yDrop;
			haveYDrop                   = true;
			}
		if ((!haveStep) && (xss->stepSet))
			{
			lzParams->step              = xss->step;
			haveStep                    = true;
			}

		if ((haveGapOpen) && (gapOpen + gapExtend < 0))
			chastise ("%s is not a valid gap open penalty\n", gapOpenStr);
		if ((haveGapExtend) && (gapExtend < 0))
			chastise ("%s is not a valid gap extension penalty\n", gapExtendStr);
		}
	else if (useUnitScores)
		{
		// use unit scoring matrix, scaled if requested

		if (!haveGapOpen)
			{
			if (scoreType == 'I')
				gapOpen = (score) ceil (unitScores_open * unitMatch);
			else
				gapOpen = (score)      (unitScores_open * unitMatch);
			haveGapOpen = true;
			}
		if (!haveGapExtend)
			{
			if (scoreType == 'I')
				gapExtend = (score) ceil (unitScores_extend * unitMatch);
			else
				gapExtend = (score)      (unitScores_extend * unitMatch);
			haveGapExtend = true;
			}
		if (!haveHspThreshold)
			{
			lzParams->hspThreshold.t = 'S';
			if (scoreType == 'I')
				lzParams->hspThreshold.s = (score) ceil (unitScores_thresh * unitMatch);
			else
				lzParams->hspThreshold.s = (score)      (unitScores_thresh * unitMatch);
			haveHspThreshold = true;
			}

		if ((haveGapOpen) && (gapOpen + gapExtend < 0))
			chastise ("%s is not a valid gap open penalty\n", gapOpenStr);
		if ((haveGapExtend) && (gapExtend < 0))
			chastise ("%s is not a valid gap extension penalty\n", gapExtendStr);

		for (r=0 ; r<4 ; r++)
				for (c=0 ; c<4 ; c++)
			myUnitScores[r][c] = (r==c)? unitMatch : unitMismatch;

		lzParams->scoring = new_dna_score_set (myUnitScores,
		                                       unitScores_X    * -unitMismatch,
		                                       unitScores_fill * -unitMismatch,
		                                       gapOpen, gapExtend);
		}
	else if (lzParams->inferScores)
		{
		; // (do nothing, lzParams->scoring will be created by inference)
		}
	else
		{
		// use blastz default scoring matrix
		if (!haveGapOpen)   gapOpen   = HOXD70_open;
		if (!haveGapExtend) gapExtend = HOXD70_extend;
		if ((haveGapOpen) && (gapOpen + gapExtend < 0))
			chastise ("%s is not a valid gap open penalty\n", gapOpenStr);
		if ((haveGapExtend) && (gapExtend < 0))
			chastise ("%s is not a valid gap extension penalty\n", gapExtendStr);
		lzParams->scoring = new_dna_score_set (HOXD70,
		                                      HOXD70_X, HOXD70_fill,
		                                      gapOpen, gapExtend);
		}

	if (haveSpecialSubs)
		{
		score worstScore = 0;

		if ((!lzParams->scoring->rowsAreDna)
		 || (!lzParams->scoring->colsAreDna))
			suicide ("scores are for quantum DNA; so special substitution scores not allowed");

		for (r=0 ; r<4 ; r++)
				for (c=0 ; c<4 ; c++)
			{
			if (specialSubScores[r][c] == worstPossibleScore)
				{
				nuc1 = bits_to_nuc[r];
				nuc2 = bits_to_nuc[c];
				specialSubScores[r][c] = lzParams->scoring->sub[nuc1][nuc2];
				}
			if (specialSubScores[r][c] < worstScore)
				worstScore = specialSubScores[r][c];
			}
		free_if_valid ("lzParams->scoring", lzParams->scoring);
		lzParams->scoring = new_dna_score_set (specialSubScores,
		                                      10*worstScore, worstScore,
		                                      gapOpen, gapExtend);
		}

	//////////
	// convert seed string to a seed structure
	//////////

	if (seedString == NULL)
		seedString = copy_string (defaultSeedString);

	parse_seeds_string (seedString, &lzParams->hitSeed, lzParams->maxIndexBits);

	if ((lzParams->hitSeed->type == 'H')
	  && (!haveWithTrans))
		lzParams->withTrans = 0;

	if ((lzParams->withTrans != 0)
	 && (lzParams->hitSeed->type != 'S')
	 && (lzParams->hitSeed->transFlips == NULL))
		chastise ("--transition can only be used with strict seeds (1s and 0s)\n");

	if (lzParams->minMatches >= 0)
		{
		if (lzParams->hitSeed->type == 'T')
			{
			if (lzParams->minMatches > lzParams->hitSeed->length)
				chastise ("--filter can't require more matches (%d) than seed (%d)\n",
						  lzParams->minMatches, lzParams->hitSeed->length);
			if (lzParams->maxTransversions + lzParams->minMatches > lzParams->hitSeed->length)
				chastise ("--filter can't require total (%d) more than seed (%d)\n",
						  lzParams->maxTransversions + lzParams->minMatches, lzParams->hitSeed->length);
			if (lzParams->maxTransversions < 0)
				chastise ("--filter can't specify negative transversions (%d)\n",
						  lzParams->maxTransversions);
			}
		else
			chastise ("--filter can only be used with half-weight seeds (Ts and 0s)\n");
		}

	pos_table_set_stat (wordWeight, lzParams->hitSeed->weight);
	pos_table_set_stat (wordSpace,  1L << lzParams->hitSeed->weight);

	lzParams->hitSeed->withTrans = lzParams->withTrans;
	seed_search_set_stat (withTrans,         lzParams->hitSeed->withTrans);
	seed_search_set_stat (minMatches,        lzParams->minMatches);
	seed_search_set_stat (maxTransversions,  lzParams->maxTransversions);

	if (debug >= 90)
		{
		print_seeds (stdout, lzParams->hitSeed);
		printf ("%s\n", seed_pattern (lzParams->hitSeed));
		}

	// set span for twins (complicated by the fact that we let the user set it
	// in terms of gap length before we know the seed length)

	if ((lzParams->noHitFiltering) && (twinsYes))
		chastise ("--rawhits can't be used with --twins\n");

	if ((lzParams->noHitFiltering) && (lzParams->gfExtend != gfexNoExtend))
		chastise ("--rawhits can't be used with --gfextend\n");

	if (twinsYes)
		{
		if (minGap <= -lzParams->hitSeed->length)
			chastise ("minGap for twins (%d) must be greater than negative of seed length (%d)\n",
			          minGap, -lzParams->hitSeed->length);
		if (maxGap < minGap)
			chastise ("maxGap for twins (%d) can't be less than min gap (%d)\n",
			          maxGap, minGap);

		lzParams->twinMinSpan = 2 * lzParams->hitSeed->length + minGap;
		lzParams->twinMaxSpan = 2 * lzParams->hitSeed->length + maxGap;
		}
	else
		lzParams->twinMinSpan = 0;  // gaps not used

	//////////
	// set up others
	//////////

	// compute default values for parameters that have not been set by the user

	if (!haveXDrop)
		{
		if (lzParams->inferScores)
			lzParams->xDrop = -1;	// (will fill in after scoring inference)
		else
			lzParams->xDrop = 10 * lzParams->scoring->sub['A'][lzParams->scoring->colChars[0]];
		}

	if (!haveYDrop)
		{
		if (lzParams->inferScores)
			lzParams->yDrop = -1;	// (will fill in after scoring inference)
		else
			lzParams->yDrop = lzParams->scoring->gapOpen + 300 * lzParams->scoring->gapExtend;
		}

	if (!haveGappedThreshold)
		lzParams->gappedThreshold = lzParams->hspThreshold;

	// make scores vs N be ambiguous, if desired and if rows and columns are
	// both DNA

	if ((nIsAmbiguous)
	 && (lzParams->scoring->colsAreDna)
	 && (lzParams->scoring->rowsAreDna))
		ambiguate_n (lzParams->scoring, 0, -2*lzParams->scoring->gapExtend);

	// create a version of the scoring set that penalized lowercase bases

	if (!lzParams->inferScores)
		{
		lzParams->maskedScoring = masked_score_set (lzParams->scoring);
		// dump_score_set (stderr, lzParams->scoring,       (u8*)"ACGTacgtNnX", (u8*)"ACGTacgtNnX");
		// dump_score_set (stderr, lzParams->maskedScoring, (u8*)"ACGTacgtNnX", (u8*)"ACGTacgtNnX");
		}

	if (lzParams->inferScores)
		maxScore = 0;  // (maxScore is not needed)
	else
		maxScore = max_in_score_matrix (lzParams->scoring);

	if (lzParams->seq2Filename != NULL) seq2Filename = lzParams->seq2Filename;
	                               else seq2Filename = "(unnamed query file)";

	if ((lzParams->seq2Filename != NULL)
	 && (strcmp_suffix(lzParams->seq2Filename,".qdna") == 0))
		lzParams->queryIsQuantum = true;

	if (lzParams->inferScores)
		{
		; // (no check here; lzParams->scoring doesn't exist yet)
		}
	else if (!lzParams->queryIsQuantum)					// DNA query
		{
		if (!lzParams->maskedScoring->colsAreDna)
			suicidef ("scores are for quantum DNA, so --quantum required");
		if ((lzParams->ballScore >= 0) || (haveBallScore))
			suicidef ("--ball can't be used with DNA (%s)", seq2Filename);
		if (codeFileName != NULL)
			suicidef ("--code can't be used with DNA (%s)", seq2Filename);
		}
	else 												// quantum DNA query
		{
		if (lzParams->maskedScoring->colsAreDna)
			suicidef ("scores are for DNA, so --quantum not allowed");
		if ((haveBallScore) && (lzParams->ballScore < 0))
			chastise (scoreFmtSimple " is not a legitimate ball threshold\n",
                      lzParams->ballScore);
		if (!haveBallScore)
			lzParams->ballScore = ballScoreFactor * maxScore
			                   * lzParams->hitSeed->weight/2;
		if (lzParams->ballScore < 0)
			suicidef ("quantum DNA (%s) requires --ball", seq2Filename);
		if ((lzParams->outputFormat == fmtAxt)
		 || (lzParams->outputFormat == fmtAxtComment))
			suicidef ("--axt doesn't support quantum DNA");
		if ((lzParams->outputFormat == fmtMaf)
		 || (lzParams->outputFormat == fmtMafComment)
		 || (lzParams->outputFormat == fmtMafNoComment))
			suicidef ("--maf doesn't support quantum DNA");
		if (codeFileName == NULL)
			lzParams->coding = NULL;
		else
			lzParams->coding = read_quantum_code_by_name (codeFileName);
		}

	// build a seed for interpolation (an n-mer exact match)

	if (lzParams->innerThreshold > 0)
		{
		char seedString[innerWordSize+1];
		int i;

		for (i=0 ; i<innerWordSize ; i++)
			seedString[i] = '1';
		seedString[innerWordSize] = 0;

		parse_seeds_string (seedString, &lzParams->innerSeed, 28);
		lzParams->innerSeed->withTrans = 0;
		}

	// decide whether we need to 'waste' time reading files just to get the
	// true sequence length

	lzParams->outputNeedsTrueLen = (lzParams->outputFormat == fmtAxt)
	                            || (lzParams->outputFormat == fmtAxtComment)
	                            || (lzParams->outputFormat == fmtMaf)
	                            || (lzParams->outputFormat == fmtMafComment)
	                            || (lzParams->outputFormat == fmtMafNoComment);

	// propagate some deep control to other modules

	if (lzParams->verbosity >= 2)
		{
		showProgress             = true;
		pos_table_showProgress   = true;
		seed_search_showProgress = true;
		}

	pos_table_verbosity      = lzParams->verbosity;
	seed_search_verbosity    = lzParams->verbosity;
	quantum_verbosity        = lzParams->verbosity;
	chain_verbosity          = lzParams->verbosity;
	gapped_extend_verbosity  = (gappedExtendVerbosity>=0)? gappedExtendVerbosity
	                                                     : lzParams->verbosity;
	tweener_verbosity        = lzParams->verbosity;
	infer_scores_verbosity   = lzParams->verbosity;

	sequences_keepFastaArrow = lzParams->lajCompatible;

	//////////
	// set up inference parameters
	//////////

	if (lzParams->inferScores)
		{
		infcontrol tempIc;

		tempIc = izParams->ic;
		*izParams = *lzParams;
		izParams->ic = tempIc;
		izParams->tracebackMem = 0;
		izParams->outputFormat = fmtNone;

		izParams->hitSeed = copy_seeds (lzParams->hitSeed);

		if (izParams->scoring == NULL)
			izParams->scoring = new_dna_score_set (unitScores,
			                                       unitScores_X, unitScores_fill,
			                                       worstPossibleScore, worstPossibleScore);

		if (infControlFileName != NULL)
			read_control_file_by_name (infControlFileName, izParams);

		if ((izParams->ic.inferScale >  0)
		 && (izParams->ic.inferScale != 1))
			scale_score_set (izParams->scoring, izParams->ic.inferScale);

		if (izParams->scoring->gapOpen == worstPossibleScore)
			{
			if (izParams->ic.inferScale <= 0)
				izParams->scoring->gapOpen =               unitScores_open;
			else if (scoreType == 'I')
				izParams->scoring->gapOpen = (score) ceil (unitScores_open * izParams->ic.inferScale);
			else
				izParams->scoring->gapOpen = (score)      (unitScores_open * izParams->ic.inferScale);
			}

		if (izParams->scoring->gapExtend == worstPossibleScore)
			{
			if (izParams->ic.inferScale <= 0)
				izParams->scoring->gapExtend =               unitScores_extend;
			else if (scoreType == 'I')
				izParams->scoring->gapExtend = (score) ceil (unitScores_extend * izParams->ic.inferScale);
			else
				izParams->scoring->gapExtend = (score)      (unitScores_extend * izParams->ic.inferScale);
			}

		izParams->maskedScoring = masked_score_set (izParams->scoring);
		}

	//////////
	// clean up
	//////////

	free_if_valid ("parse_options (seedString)",         seedString);
	free_if_valid ("parse_options (scoreFileName)",      scoreFileName);
	free_if_valid ("parse_options (codeFileName)",       codeFileName);
	free_if_valid ("parse_options (infControlFileName)", infControlFileName);
	}


static void print_params
   (FILE*	f)
	{
	if (lzParams.seq1 != NULL) fprintf (f, "seq 1: %s\n", lzParams.seq1->fileName);
	if (lzParams.seq2 != NULL) fprintf (f, "seq 2: %s\n", lzParams.seq2->fileName);
	if (lzParams.selfCompare)   fprintf (f, "--self\n");
	print_score_matrix (f, lzParams.scoring, true);
	if      (lzParams.whichStrand > 0) fprintf (f, "--both\n");
	else if (lzParams.whichStrand < 0) fprintf (f, "--minus\n");
	else                             fprintf (f, "--plus\n");
	fprintf (f, "--step=%u\n", lzParams.step);
	fprintf (f, "--seed=%s\n", seed_pattern(lzParams.hitSeed));
	if (lzParams.gfExtend == gfexXDrop) fprintf (f, "--gfextend\n");
	fprintf (f, "--hspthresh=%s\n",    score_thresh_to_string (&lzParams.hspThreshold));
	fprintf (f, "--gappedthresh=%s\n", score_thresh_to_string (&lzParams.gappedThreshold));
	fprintf (f, "--xDrop=" scoreFmtSimple "\n", lzParams.xDrop);
	fprintf (f, "--yDrop=" scoreFmtSimple "\n", lzParams.yDrop);
	fprintf (f, "%s\n", (lzParams.entropicHsp)? "--entropy" : "--noentropy");
	if (lzParams.minMatches >= 0) fprintf (f, "--filter=%d:%d\n", lzParams.minMatches, lzParams.maxTransversions);
	if (lzParams.twinMinSpan > 0) fprintf (f, "--twins=%d:%d\n", lzParams.twinMinSpan-2*lzParams.hitSeed->length, lzParams.twinMaxSpan-2*lzParams.hitSeed->length);
	if (lzParams.innerThreshold > 0) fprintf (f, "--innerthresh=" scoreFmtSimple "\n", lzParams.innerThreshold);
	if (lzParams.tracebackMem != defaultParams.tracebackMem) fprintf (f, "--traceback=%d\n", lzParams.tracebackMem);
	}

//----------
//
// read_control_file_by_name, read_control_file--
//	Read control data from a file (see format description below).
//
//----------
//
// Arguments:
//	FILE*		f:		(read_control_file only) The file that control data is
//						.. to be read from.  This should already be open for
//						.. text read.
//	char*		name:	The name of the file that control data is to be read
//						.. from.  For read_control_file this is only used for
//						.. reporting problems to the user (and may be NULL).
//	control*	params:	Control data to fill in.
//
// Returns:
//	(nothing)
//
//----------
//
// Control Data File Format
// ========================
//
// Here's an example:
//
//		min_identity  = 25.0%	# 25th percentile
//		max_identity  = 75.0%	# 75th percentile
//		hsp_threshold = 3000
//
// The control data consists of name-value settings.  Valid names are as
// follows:
//
//		min_identity		The range of sequence identity upon which inference
//		max_identity		.. is based.  Only alignment blocks within this
//							.. range contribute to inference.  If the value ends
//							.. with a percent sign, the range is a percentile of
//							.. the values found in the overall alignment (other-
//							.. wise it is a fixed percentage.
//
//		min_coverage		The range of query coverage upon which inference
//		max_coverage		.. is based.  Only alignment blocks within this
//							.. range (as a percentage of the query sequence)
//							.. contribute to inference.
//
//		inference_scale		The value for the largest substitution score (i.e.
//							.. the score for the best match).  All other scores
//							.. are scaled by the same factor.  If this is an
//							.. integer (i.e. has no decimal point), then all
//							.. scores will be rounded to an integer as well.
//
//		hsp_threshold		These correspond to the command line --hspthresh
//		gapped_threshold	.. and --gappedthresh options (also known as K and L
//							.. in BLASTZ lingo).  They can be specified as
//							.. mulitples of the scale, e.g.
//							..    hsp_threshold = 20*inference_scale
//							.. Further, the gapped threshold can be specified
//							.. as a multiple of the hsp threshold, e.g.
//							..    gapped_threshold = 1.2*hsp_threshold
//
//		max_sub_iterations	Limits on the number of iterations that will be
//		max_gap_iterations	.. performed.  For example,
//							..    max_sub_iterations = 1
//							..    max_gap_iterations = 0
//							.. will just do one pass and create a substitution
//							.. scoring matrix.
//
//		gap_open_penalty	Correspond to the command line --gap=<[open,]extend>
//		gap_extend_penalty	.. option (also known as O and E in BLASTZ lingo).
//							.. These are the values used for the first iteration
//							.. of gap-scoring inference.  They can be specified
//							.. as mulitples of the scale, and the extend penalty
//							.. can be a multiple of the open penalty.
//
//		step				Corresponds to the command line --step option (also
//							.. known as Z in BLASTZ lingo).
//
//		entropy				Corresponds to the command line --entropy option
//							.. (also known as P in BLASTZ lingo).  Legal values
//							.. are "on" or "off".
//
//----------

static void read_control_file_by_name
   (char*		name,
	control*	params)
	{
	FILE*		f;

	if (name == NULL)
		suicide ("can't open NULL file in read_control_file_by_name()");

	f = fopen_or_die (name, "rt");
	read_control_file (f, name, params);
	fclose_if_valid (f);
	}

static void read_control_file
   (FILE*		f,
	char*		_name,
	control*	params)
	{
	char		line[201];
	char*		name = _name;
	int			lineNum, len, valLen, missingEol;
	char*		waffle;
	char*		valString = NULL;
	int			idIsPercentile = -1;
	int			haveMinIdentity = false;
	int			haveMaxIdentity = false;
	int			haveMinCoverage = false;
	int			haveMaxCoverage = false;
	int			tempInt;

	if (name == NULL)
		name = "(unnamed file)";

	//////////
	// read assignments
	//////////

	lineNum    = 0;
	missingEol = false;

	while (fgets (line, sizeof(line), f) != NULL)
		{
		lineNum++;

		// check for lines getting split by fgets (the final line in the file
		// might not have a newline, but no internal lines can be that way)

		if (missingEol)
			goto line_too_long;

		len = strlen (line);
		if (len == 0) continue;
		missingEol = (line[len-1] != '\n');

		// trim blanks, end of line, and comments, and ignore blank lines

		if (line[len-1] == '\n') line[--len] = 0;

		waffle = strchr (line, '#');
		if (waffle != NULL) *waffle = 0;

		trim_string (line);
		if (line[0] == 0) continue;

		// separate the value from the assignment

		valString = strchr (line, '=');
		if (valString == NULL)
			goto invalid_line;

		*(valString++) = 0;
		trim_string (line);
		trim_string (valString);
		valLen = strlen (valString);
		if (valLen == 0)
			goto empty_assignment;

		// parse the assignment

		// inference_scale

		if (strcmp (line, "inference_scale") == 0)
			{
#if (scoreType != 'I')
			int  v;
			char extra;
#endif

			if (strcmp (valString, "none") == 0)
				{
				params->ic.inferScale = 0;
				params->ic.writeAsInt = false;
				}
			else
				{
				params->ic.inferScale = string_to_score (valString);

#if (scoreType == 'I')
				params->ic.writeAsInt = false; // (no need to force it to an int)
#else
				params->ic.writeAsInt = (sscanf (valString, "%d%c", &v, &extra) == 1);
#endif
				}
			continue;
			}

		// hsp_threshold, gapped_threshold
		// $$$ the use of expressions involving other settings should be
		// $$$ .. expanded and generalized

		if (strcmp (line, "hsp_threshold") == 0)
			{
			params->ic.hspThresholdIsRatio = ratioNone;
			if (strcmp_prefix (valString, "top") == 0)
				params->hspThreshold = string_to_score_thresh (valString);
			else if (strcmp_suffix (valString,"*inference_scale") == 0)
				{
				valString[valLen-strlen("*inference_scale")] = 0;
				params->hspThreshold.t = 'S';
				params->hspThreshold.s = string_to_double (valString);
				if (params->ic.inferScale > 0)
					params->hspThreshold.s *= params->ic.inferScale;
				else
					params->ic.hspThresholdIsRatio = ratioMaxSubScore;
				}
			else if (strcmp_suffix (valString,"*worst_substitution") == 0)
				{
				valString[valLen-strlen("*worst_substitution")] = 0;
				params->hspThreshold.t = 'S';
				params->hspThreshold.s = string_to_double (valString);
				params->ic.hspThresholdIsRatio = ratioMinSubScore;
				}
			else
				{
				params->hspThreshold.t = 'S';
				params->hspThreshold.s = string_to_score (valString);
				}
			continue;
			}

		if (strcmp (line, "gapped_threshold") == 0)
			{
			params->ic.gappedThresholdIsRatio = ratioNone;
			if (strcmp_prefix (valString, "top") == 0)
				params->gappedThreshold = string_to_score_thresh (valString);
			else if (strcmp_suffix (valString,"*inference_scale") == 0)
				{
				valString[valLen-strlen("*inference_scale")] = 0;
				params->gappedThreshold.t = 'S';
				params->gappedThreshold.s = string_to_double (valString);
				if (params->ic.inferScale > 0)
					params->gappedThreshold.s *= params->ic.inferScale;
				else
					params->ic.gappedThresholdIsRatio = ratioMaxSubScore;
				}
			else if (strcmp_suffix (valString,"*worst_substitution") == 0)
				{
				valString[valLen-strlen("*worst_substitution")] = 0;
				params->gappedThreshold.t = 'S';
				params->gappedThreshold.s = string_to_double (valString);
				params->ic.gappedThresholdIsRatio = ratioMinSubScore;
				}
			else if (strcmp (valString, "hsp_threshold") == 0)
				params->gappedThreshold = params->hspThreshold;
			else
				{
				params->gappedThreshold.t = 'S';
				params->gappedThreshold.s = string_to_score (valString);
				}
			continue;
			}

		// gap_open_penalty, gap_extend_penalty

		if (strcmp (line, "gap_open_penalty") == 0)
			{
			params->ic.gapOpenIsRatio = ratioNone;
			if (strcmp_suffix (valString,"*inference_scale") == 0)
				{
				valString[valLen-strlen("*inference_scale")] = 0;
				params->scoring->gapOpen = string_to_double (valString);
				if (params->ic.inferScale > 0)
					params->scoring->gapOpen *= params->ic.inferScale;
				else
					params->ic.gapOpenIsRatio = ratioMaxSubScore;
				}
			else if (strcmp_suffix (valString,"*worst_substitution") == 0)
				{
				valString[valLen-strlen("*worst_substitution")] = 0;
				params->scoring->gapOpen = string_to_double (valString);
				params->ic.gapOpenIsRatio = ratioMinSubScore;
				}
			else
				params->scoring->gapOpen = string_to_score (valString);
			continue;
			}

		if (strcmp (line, "gap_extend_penalty") == 0)
			{
			params->ic.gapExtendIsRatio = ratioNone;
			if (strcmp_suffix (valString,"*inference_scale") == 0)
				{
				valString[valLen-strlen("*inference_scale")] = 0;
				params->scoring->gapExtend = string_to_double (valString);
				if (params->ic.inferScale > 0)
					params->scoring->gapExtend *= params->ic.inferScale;
				else
					params->ic.gapExtendIsRatio = ratioMaxSubScore;
				}
			else if (strcmp_suffix (valString,"*worst_substitution") == 0)
				{
				valString[valLen-strlen("*worst_substitution")] = 0;
				params->scoring->gapExtend = string_to_double (valString);
				params->ic.gapExtendIsRatio = ratioMinSubScore;
				}
			else if (strcmp_suffix (valString,"*gap_open_penalty") == 0)
				{
				valString[valLen-strlen("*gap_open_penalty")] = 0;
				params->scoring->gapExtend = string_to_double (valString)
				                           * params->scoring->gapOpen;
				params->ic.gapExtendIsRatio = params->ic.gapOpenIsRatio;
				}
			else
				params->scoring->gapExtend = string_to_score (valString);
			continue;
			}

		// entropy

		if (strcmp (line, "entropy") == 0)
			{
			if (strcmp (valString, "on") == 0)
				params->entropicHsp = true;
			else if (strcmp (valString, "off") == 0)
				params->entropicHsp = false;
			else
				goto on_off_mismatch;
			continue;
			}

		// max_sub_iterations, max_gap_iterations

		if (strcmp (line, "max_sub_iterations") == 0)
			{
			params->ic.subIterations = string_to_int (valString);
			continue;
			}

		if (strcmp (line, "max_gap_iterations") == 0)
			{
			params->ic.gapIterations = string_to_int (valString);
			continue;
			}

		// step

		if (strcmp (line, "step") == 0)
			{
			tempInt = string_to_int (valString);
			if (tempInt <= 0)
				goto bad_step;
			params->step = tempInt;
			continue;
			}

		// min_identity, max_identity
		//
		// Ideally the user will set both of these;  however, if one is set but
		// not the other, we peg the other to the edge of the range;  further,
		// we validate that the user specifies percentile or non-percentile the
		// same for both

		if ((strcmp (line, "min_identity") == 0)
		 || (strcmp (line, "max_identity") == 0))
			{
			if (valString[valLen-1] == '%')
				{
				valString[--valLen] = 0;
				if (idIsPercentile == false)
					goto percentile_mismatch;
				if (idIsPercentile == -1)
					params->ic.idIsPercentile = idIsPercentile = true;
				}
			else if (idIsPercentile == true)
				goto percentile_mismatch;
			else if (idIsPercentile == -1)
				params->ic.idIsPercentile = idIsPercentile = false;

			if (strcmp (line, "min_identity") == 0)
				{
				params->minIdentity = string_to_double (valString) / 100;
				haveMinIdentity = true;
				if (!haveMaxIdentity) params->maxIdentity = 1.0;
				}
			else
				{
				params->maxIdentity = string_to_double (valString) / 100;
				haveMaxIdentity = true;
				if (!haveMinIdentity) params->minIdentity = 0.0;
				}
			continue;
			}

		// min_coverage, max_coverage
		//
		// Ideally the user will set both of these;  however, if one is set but
		// not the other, we peg the other to the edge of the range

		if ((strcmp (line, "min_coverage") == 0)
		 || (strcmp (line, "max_coverage") == 0))
			{
			if (strcmp (line, "min_coverage") == 0)
				{
				params->minCoverage = string_to_double (valString) / 100;
				haveMinCoverage = true;
				if (!haveMaxCoverage) params->maxCoverage = 1.0;
				}
			else
				{
				params->maxCoverage = string_to_double (valString) / 100;
				haveMaxCoverage = true;
				if (!haveMinCoverage) params->minCoverage = 0.0;
				}
			continue;
			}

		goto unknown_assignment;
		}

	return;

	//////////
	// failure exits
	//////////

line_too_long:
	suicidef ("line is too long (%s: line %d)", name, lineNum-1);

invalid_line:
	suicidef ("invalid line (%s: line %d) %s", name, lineNum, line);

empty_assignment:
	suicidef ("empty value in assignment (%s: line %d) %s=",
			  name, lineNum, line);

percentile_mismatch:
	suicidef ("assignment of identity percentile mismatches earlier setting"
	          " (%s: line %d) %s=",
	          name, lineNum, line);

on_off_mismatch:
	suicidef ("invalid on/off in assignment (%s: line %d) %s=%s",
	          name, lineNum, line, valString);

unknown_assignment:
	suicidef ("invalid name in assignment (%s: line %d) %s=%s",
	          name, lineNum, line, valString);

bad_step:
	suicidef ("invalid value for step (%s: line %d) %s=%s",
	          name, lineNum, line, valString);
	}

//----------
//
// print_job_header, print_job_footer, print_header, print_align_list,
// print_match--
//
//----------

void print_job_header (void)
	{
	int outputFormat = currParams->outputFormat;
	char* progName, _progName[101];
	int n;

	n = snprintf (NULL, 0, "%s.v%s.%s.%s",
	              programName, programVersionMajor, programVersionMinor, programVersionSubMinor);
	if (((unsigned) n) < sizeof(_progName))
		{
		sprintf (_progName,
		         "%s.v%s.%s.%s",
		         programName, programVersionMajor, programVersionMinor, programVersionSubMinor);
		progName = _progName;
		}
	else
		progName = programName;

	switch (outputFormat)
		{
		case fmtGfa:
		case fmtGfaNoScore:
			print_gfa_job_header
			   (stdout, progName, currParams->seq1->fileName, currParams->seq2->fileName);
			break;
		case fmtLav:
		case fmtLavComment:
		case fmtLavScore:
		case fmtLavText:
		case fmtLavInfScores:
			print_lav_job_header
			   (stdout, progName,
			    currParams->seq1->fileName, currParams->seq2->fileName, currParams->args,
		        currParams->scoring, &currParams->hspThreshold, &currParams->gappedThreshold);
			if (outputFormat == fmtLavText)
				goto text_format;
			if (outputFormat == fmtLavInfScores)
				goto inf_scores_format;
			break;
		case fmtAxt:
		case fmtAxtComment:
			print_axt_job_header
			   (stdout, progName, currParams->args,
		        currParams->scoring,
		        &currParams->hspThreshold, &currParams->gappedThreshold,
		        currParams->xDrop, currParams->yDrop);
			break;
		case fmtMaf:
		case fmtMafComment:
		case fmtMafNoComment:
			print_maf_job_header
			   (stdout, progName, currParams->args,
		        currParams->scoring,
		        &currParams->hspThreshold, &currParams->gappedThreshold,
		        currParams->xDrop, currParams->yDrop,
		        (outputFormat != fmtMafNoComment));
			break;
		case fmtText:
		text_format:
			print_text_align_job_header
			   (stdout, progName, currParams->seq1->fileName, currParams->seq2->fileName);
			break;
		case fmtInfStats:
			init_inference_stats_job (currParams->seq1, currParams->seq2);
			break;
		case fmtIdDist:
			init_identity_dist_job (currParams->seq1, currParams->seq2);
			break;
		case fmtInfScores:
		inf_scores_format:
			; // (do nothing)
			break;
		case fmtHspComp:
		case fmtDeseed:
		case fmtNone:
			; // (do nothing)
			break;
		}
	}

static void print_quantum_coding (qcode* coding)
	{
	int outputFormat = currParams->outputFormat;

	if ((outputFormat == fmtText)
	 || (outputFormat == fmtLavText))
		print_text_align_quantum_coding (coding);
	}

static void print_job_footer (void)
	{
	int outputFormat = currParams->outputFormat;

	switch (outputFormat)
		{
		case fmtGfa:
		case fmtGfaNoScore:
			print_gfa_job_footer (stdout);
			break;
		case fmtLav:
		case fmtLavComment:
		case fmtLavScore:
		case fmtLavText:
		case fmtLavInfScores:
			print_lav_job_footer (stdout);
			if (outputFormat == fmtLavText)
				goto text_format;
			if (outputFormat == fmtLavInfScores)
				goto inf_scores_format;
			break;
		case fmtAxt:
		case fmtAxtComment:
			print_axt_job_footer (stdout);
			break;
		case fmtMaf:
		case fmtMafComment:
		case fmtMafNoComment:
			print_maf_job_footer (stdout);
			break;
		case fmtText:
		text_format:
			print_text_align_job_footer (stdout);
			break;
		case fmtInfStats:
			print_inference_stats_job (stdout);
			break;
		case fmtIdDist:
			print_identity_dist_job (stdout);
			break;
		case fmtInfScores:
		inf_scores_format:
			; // (do nothing)
			break;
		case fmtHspComp:
		case fmtDeseed:
		case fmtNone:
			; // (do nothing)
			break;
		}
	}

static void print_header (void)
	{
	int outputFormat = currParams->outputFormat;

	switch (outputFormat)
		{
		case fmtGfa:
		case fmtGfaNoScore:
			print_gfa_header (stdout, currParams->seq1, currParams->seq2);
			break;
		case fmtLav:
		case fmtLavComment:
		case fmtLavScore:
		case fmtLavText:
		case fmtLavInfScores:
			print_lav_header (stdout, currParams->seq1, currParams->seq2);
			if (outputFormat == fmtLavText)
				goto text_format;
			if (outputFormat == fmtLavInfScores)
				goto inf_scores_format;
			break;
		case fmtAxt:
		case fmtAxtComment:
			print_axt_header (stdout, currParams->seq1, currParams->seq2);
			break;
		case fmtMaf:
		case fmtMafComment:
		case fmtMafNoComment:
			print_maf_header (stdout, currParams->seq1, currParams->seq2);
			break;
		case fmtText:
		text_format:
			print_text_align_header (stdout, currParams->seq1, currParams->seq2);
			break;
		case fmtHspComp:
		case fmtInfStats:
		case fmtInfScores:
		inf_scores_format:
			; // (do nothing)
			break;
		case fmtIdDist:
		case fmtDeseed:
		case fmtNone:
			; // (do nothing)
			break;
		}
	}

static void print_align_list (alignel* alignList)
	{
	int outputFormat = currParams->outputFormat;
	alignel* a;

	if (!headerPrinted)
		{ print_header ();  headerPrinted = true; }

	if (infer_scores_dbgShowIdentity)
		{
		unspos numer, denom;
		u32    bin;

		for (a=alignList ; a!=NULL ; a=a->next)
			{
			alignment_identity (currParams->seq1, currParams->seq2, a,
			                    &numer, &denom);
			bin = identity_bin (numer, denom);
			// nota bene: positions written as 1-based
			print_generic (stdout, unsposSlashFmt " pct_identity=" unsposSlashFmt " (bin as " identityBinFormat ")",
			                       a->beg1, a->beg2, numer, denom, bin_to_identity (bin));
			}
		}

	switch (outputFormat)
		{
		case fmtGfa:
		case fmtGfaNoScore:
			print_gfa_align_list (stdout,
			                      (outputFormat == fmtGfa)? currParams->scoring
			                                              : NULL,
			                      alignList,
								  currParams->seq1, currParams->seq2);
			break;
		case fmtLav:
		case fmtLavComment:
		case fmtLavScore:
		case fmtLavInfScores:
			print_lav_align_list (stdout,
			                      alignList,
			                      currParams->seq1, currParams->seq2);
			if (outputFormat == fmtLavInfScores)
				goto inf_scores_format;
			break;
		case fmtLavText:
			for (a=alignList ; a!=NULL ; a=a->next)
				{
				print_lav_align        (stdout,
										a->seq1, a->beg1-1, a->end1,
										a->seq2, a->beg2-1, a->end2,
										a->script, a->s);
				print_text_align_align (stdout,
										currParams->seq1, a->beg1-1, a->end1,
										currParams->seq2, a->beg2-1, a->end2,
										a->script, a->s,
										currParams->textContext);
				}
			break;
		case fmtAxt:
		case fmtAxtComment:
			print_axt_align_list (stdout, alignList,
			                      currParams->seq1, currParams->seq2,
			                      /* comments */ outputFormat==fmtAxtComment);
			break;
		case fmtMaf:
		case fmtMafNoComment:
			print_maf_align_list (stdout,
			                      alignList, currParams->seq1, currParams->seq2,
			                      /* comments */ false);
			break;
		case fmtMafComment:
			print_maf_align_list (stdout,
			                      alignList, currParams->seq1, currParams->seq2,
			                      /* comments */ true);
			break;
		case fmtText:
			print_text_align_align_list (stdout,
			                             alignList, currParams->seq1, currParams->seq2,
			                             currParams->textContext);
			break;
		case fmtInfStats:
			infer_stats_from_align_list (alignList, currParams->seq1, currParams->seq2);
			break;
		case fmtInfScores:
		inf_scores_format:
			gather_stats_from_align_list (alignList, currParams->seq1, currParams->seq2);
			break;
		case fmtIdDist:
			identity_dist_from_align_list (alignList, currParams->seq1, currParams->seq2);
			break;
		case fmtHspComp:
		case fmtDeseed:
		case fmtNone:
			; // (do nothing)
			break;
		}
	}

static void print_match (unspos pos1, unspos pos2, unspos length, score s)
	// pos1 and pos2 are the positions of first character in the match,
	// .. (origin-0).
	{
	int outputFormat = currParams->outputFormat;

	if (!headerPrinted)
		{ print_header ();  headerPrinted = true; }

	if (infer_scores_dbgShowIdentity)
		{
		unspos numer, denom;
		u32    bin;

		segment_identity (currParams->seq1, pos1, currParams->seq2, pos2, length,
		                  &numer, &denom);
		bin = identity_bin (numer, denom);
		// nota bene: positions written as 1-based
		print_generic (stdout, unsposSlashFmt " pct_identity=" unsposSlashFmt
		                       " (bin as " identityBinFormat ")",
		                       pos1+1, pos2+1, numer, denom,
		                       bin_to_identity (bin));
		}

	switch (outputFormat)
		{
		case fmtGfa:
		case fmtGfaNoScore:
			print_gfa_match (stdout,
			                 currParams->seq1, pos1,
			                 currParams->seq2, pos2, length,
		                     (outputFormat == fmtGfa)? s : 0);
			break;
		case fmtLav:
		case fmtLavComment:
		case fmtLavText:
		case fmtLavInfScores:
			print_lav_match (stdout,
			                 currParams->seq1, pos1,
			                 currParams->seq2, pos2, length,
			                 s);
			if (outputFormat == fmtLavText)
				goto text_format;
			if (outputFormat == fmtLavInfScores)
				goto inf_scores_format;
			break;
		case fmtLavScore:
			print_lavscore_match (stdout,
			                      currParams->seq1, pos1,
			                      currParams->seq2, pos2, length,
			                      s);
			break;
		case fmtAxt:
		case fmtAxtComment:
			print_axt_match (stdout,
			                 currParams->seq1, pos1,
			                 currParams->seq2, pos2, length,
			                 s, /* comments */ outputFormat==fmtAxtComment);
			break;
		case fmtMaf:
		case fmtMafNoComment:
			print_maf_match (stdout,
			                 currParams->seq1, pos1,
			                 currParams->seq2, pos2, length,
			                 s, /* comments */ false);
			break;
		case fmtMafComment:
			print_maf_match (stdout,
			                 currParams->seq1, pos1,
			                 currParams->seq2, pos2, length,
			                 s, /* comments */ true);
			break;
		case fmtText:
		text_format:
			print_text_align_match (stdout,
			                        currParams->seq1, pos1,
			                        currParams->seq2, pos2, length,
			                        s, currParams->textContext);
			break;
		case fmtHspComp:
			print_match_composition (stdout,
			                         currParams->seq1, pos1,
			                         currParams->seq2, pos2, length,
			                         s, currParams->hitSeed, currParams->step);
			break;
		case fmtInfStats:
			infer_stats_from_match (currParams->seq1, pos1,
			                        currParams->seq2, pos2, length);
			break;
		case fmtInfScores:
		inf_scores_format:
			gather_stats_from_match (currParams->seq1, pos1,
			                         currParams->seq2, pos2, length);
			break;
		case fmtIdDist:
			identity_dist_from_match (currParams->seq1, pos1,
			                          currParams->seq2, pos2, length);
			break;
		case fmtDeseed:
			dump_match (stdout,
			            currParams->seq1, pos1,
			            currParams->seq2, pos2, length);
			printf ("\n");
			break;
		case fmtNone:
			; // (do nothing)
			break;
		}
	}

static void print_m_stanza (census* cen)
	{ // note that census might be NULL
	int outputFormat = currParams->outputFormat;

	switch (outputFormat)
		{
		case fmtLav:
		case fmtLavComment:
		case fmtLavScore:
		case fmtLavText:
			print_lav_m_stanza (stdout, cen);
			break;
		case fmtGfa:
		case fmtGfaNoScore:
		case fmtAxt:
		case fmtAxtComment:
		case fmtMaf:
		case fmtMafComment:
		case fmtMafNoComment:
		case fmtText:
		case fmtHspComp:
		case fmtInfStats:
		case fmtInfScores:
		case fmtIdDist:
		case fmtDeseed:
		case fmtNone:
			; // (do nothing)
			break;
		}
	}

static void print_census_stanza (census* cen)
	{
	int outputFormat = currParams->outputFormat;

	switch (outputFormat)
		{
		case fmtLav:
		case fmtLavComment:
		case fmtLavScore:
		case fmtLavText:
			print_lav_census_stanza (stdout, cen);
			break;
		case fmtGfa:
		case fmtGfaNoScore:
		case fmtAxt:
		case fmtAxtComment:
		case fmtMaf:
		case fmtMafComment:
		case fmtMafNoComment:
		case fmtText:
		case fmtHspComp:
		case fmtInfStats:
		case fmtInfScores:
		case fmtIdDist:
		case fmtDeseed:
		case fmtNone:
			; // (do nothing)
			break;
		}
	}

static void print_x_stanza (unspos numMasked)
	{
	int outputFormat = currParams->outputFormat;

	switch (outputFormat)
		{
		case fmtGfa:
		case fmtGfaNoScore:
			print_gfa_generic (stdout, 'x', "num_masked=" unsposFmt, numMasked);
			break;
		case fmtLav:
		case fmtLavComment:
		case fmtLavScore:
		case fmtLavText:
			print_lav_x_stanza (stdout, numMasked);
			break;
		case fmtAxt:
		case fmtAxtComment:
		case fmtMaf:
		case fmtMafComment:
		case fmtMafNoComment:
		case fmtText:
		case fmtHspComp:
		case fmtInfStats:
		case fmtIdDist:
		case fmtDeseed:
			print_generic (stdout, "num_masked=" unsposFmt, numMasked);
			break;
		case fmtInfScores:
		case fmtNone:
			; // (do nothing)
			break;
		}
	}

static void print_generic
   (FILE*		f,
	const char*	format,
	...)
	{
	int outputFormat;
	va_list	args;

	va_start (args, format);

	outputFormat = currParams->outputFormat;

	switch (outputFormat)
		{
		case fmtGfa:
		case fmtGfaNoScore:
			vprint_gfa_generic (f, 'z', format, args);
			break;
		case fmtLavComment:
			vprint_lav_comment (f, format, args);
			break;
		case fmtLavText:
			vprint_lav_comment (f, format, args);
			if (format != NULL)
				{
				va_end (args);
				va_start (args, format);
				vfprintf (f, format, args);
				fprintf  (f, "\n");
				}
			break;
		case fmtAxtComment:
			vprint_axt_comment (f, format, args);
			break;
		case fmtMafComment:
			vprint_maf_comment (f, format, args);
			break;
		case fmtText:
			if (format != NULL)
				{
				vfprintf (f, format, args);
				fprintf  (f, "\n");
				}
			break;
		case fmtLav:
		case fmtLavScore:
		case fmtAxt:
		case fmtMaf:
		case fmtMafNoComment:
		case fmtHspComp:
		case fmtInfStats:
		case fmtInfScores:
		case fmtIdDist:
		case fmtDeseed:
		case fmtNone:
			; // (do nothing)
			break;
		}

	va_end (args);
	}

static void print_options
   (void)
	{
	print_generic (stdout, "seed=%s%s",
	                       seed_pattern(currParams->hitSeed),
	                       (currParams->hitSeed->withTrans == 0)?"":
	                       (currParams->hitSeed->withTrans == 1)?" w/transition"
	                                                            :" w/2 transitions");
	//print_generic (stdout, "--hspthresh=" scoreFmtSimple, currParams->hspThreshold);
	//print_generic (stdout, "--gappedthresh=" scoreFmtSimple, currParams->gappedThreshold);
	//print_generic (stdout, "--xDrop=" scoreFmtSimple, currParams->xDrop);
	//print_generic (stdout, "--yDrop=" scoreFmtSimple, currParams->yDrop);
	//print_generic (stdout, "%s", (currParams->entropicHsp)? "--entropy" : "--noentropy");
	//if (currParams->minMatches >= 0) print_generic (stdout, 'z', "--filter=%d:%d", currParams->minMatches, currParams->maxTransversions);
	if (currParams->twinMinSpan > 0) print_generic (stdout, "twins=%d:%d", currParams->twinMinSpan-2*currParams->hitSeed->length, currParams->twinMaxSpan-2*currParams->hitSeed->length);
	                        else print_generic (stdout, "notwins");
	print_generic (stdout, "step=%u", currParams->step);
	}

//----------
//
// print_match_composition--
//	Print a gap-free alignment including position and composition (counts of
//	matched dna letter pairs).
//
// Typical output is shown below, with a header added.  The first letter of the
// pairs is from sequence 1, the second from sequence 2.  P is the 'discovery
// probability'-- the probability that this HSP would be discovered for this
// (seed,Z) combination, over random sequence positions.
//
//	id score pos1/pos2  len p    AA AC AG AT CA CC CG CT GA GC GG GT TA TC TG TT
//	92 121  1475+/1395- 145 .750 27  0  4  0  1 38  2  0  3  0 45  0  1  1  0 23
//	88  28  3374+/4837-  42 .200 18  0  0  0  1  7  0  0  2  1  7  1  0  0  0  5
//	 ...
//
//----------
//
// Arguments:
//	FILE*	f:		The file to print to.
//	seq*	seq1:	One sequence.
//	unspos	pos1:	The first aligned position in sequence 1.
//	seq*	seq2:	The other sequence.
//	unspos	pos2:	The first aligned position in sequence 2.
//	unspos	length:	The length of the alignment.
//	seed*	hitSeed: Seeding strategy for the hits that found this match.
//	u32		step:	Positional step size in the search for those hits.
//
// Returns:
//	(nothing)
//
//----------

static void print_match_composition
   (FILE*	f,
	seq*	seq1,
	unspos	pos1,
	seq*	seq2,
	unspos	pos2,
	unspos	length,
	score	s,
	seed*	hitSeed,
	u32		step)
	{
	int		pctId;
	unspos	count[4][4];
	float	p;
	char	pstr[6];
	int		ix, iy;

	// compute percent identity, match compostion, and discovery probability

	pctId = percent_identical (seq1, pos1, seq2, pos2, length);
	match_composition (seq1, pos1, seq2, pos2, length, count);
	p = discovery_probability (seq1, pos1+length, seq2, pos2+length, length,
	                           hitSeed, step);

	// convert discovery probability to a string

	if      (p < 0.0) p = 0.0;
	else if (p > 1.0) p = 1.0;

	snprintf (pstr, sizeof(pstr), "%.3f", p);
	if (pstr[0] == '1') // (1.000 -> 1.00)
		pstr[4] = 0;
	else				// (0.XXX -> .XXX)
		{
		pstr[0] = pstr[1];
		pstr[1] = pstr[2];
		pstr[2] = pstr[3];
		pstr[3] = pstr[4];
		pstr[4] = 0;
		}

	// print it

	fprintf (f, "%d " scoreFmtSimple " " unsposSlashSFmt " " unsposFmt " %s",
	            pctId, s,
	            pos1+1, ((seq1->revCompFlags & rcf_rev) != 0)? "-" : "+",
	            pos2+1, ((seq2->revCompFlags & rcf_rev) != 0)? "-" : "+",
	            length, pstr);

	for (ix=0 ; ix<4 ; ix++)
		for (iy=0 ; iy<4 ; iy++)
			fprintf (f, " " unsposFmt, count[ix][iy]);

	fprintf (f, "\n");
	}

//----------
//
// dump_match--
//	Dump the nucleotides (from each sequence) for a gap-free alignment.
//
//----------
//
// Arguments:
//	FILE*	f:		The file to print to.
//	seq*	seq1:	One sequence.
//	unspos	pos1:	The first aligned position in sequence 1.
//	seq*	seq2:	The other sequence.
//	unspos	pos2:	The first aligned position in sequence 2.
//	unspos	length:	The length of the alignment.
//
// Returns:
//	(nothing)
//
//----------

static void dump_match
   (FILE*	f,
	seq*	seq1,
	unspos	pos1,
	seq*	seq2,
	unspos	pos2,
	unspos	length)
	{
	print_prefix (f, (char*) seq1->v + pos1, length);
	fprintf      (f, "\n");
	print_prefix (f, (char*) seq2->v + pos2, length);
	fprintf      (f, "\n");
	}

//----------
//
// lastz_zero_stats--
//	Clear the statistics for this module.
//
//----------
//
// Arguments:
//	(none)
//
// Returns:
//	(nothing)
//
//----------

void lastz_zero_stats
   (void)
	{
#ifdef collect_stats

	// set 'em en masse to zero

	memset (&lastzStats, 0, sizeof(lastzStats));

	// set any values that might be floating point to zero (fp bit pattern for
	// zero may not be all-bits-zero)

	// (none to set, yet)

#endif // collect_stats
	}

//----------
//
// lastz_show_stats_before, lastz_show_stats--
//	Show the statistics that have been collected for this module.
//
//----------
//
// Arguments:
//	FILE*		f:	The file to print the stats to.
//
// Returns:
//	(nothing)
//
//----------

static void lastz_show_stats_before
   (arg_dont_complain(FILE* f))
	{
#ifdef collect_stats
	if (f == NULL) return;
	fprintf (f, "-------------------\n");
	fprintf (f, "     target length: %s\n",     commatize (lzParams.seq1->len));
	if (lzParams.seq2 != NULL)
		fprintf (f, "      query length: %s\n", commatize (lzParams.seq2->len));
	fprintf (f, "         step size: %u\n",     currParams->step);
	fprintf (f, "-------------------\n");
#endif // collect_stats
	}


static void lastz_show_stats
   (arg_dont_complain(FILE* f))
	{
#ifdef collect_stats
	if (f == NULL) return;
	fprintf (f, "          run time: %.3f seconds\n", lastzStats.runTime);
	fprintf (f, "-------------------\n");
#endif // collect_stats
	}

