//-------+---------+---------+---------+---------+---------+---------+--------=
//
// File: lastz.h
//
//----------

#ifndef lastz_H					// (prevent multiple inclusion)
#define lastz_H

// other files

#include "dna_utilities.h"		// dna/scoring stuff
#include "seeds.h"				// seed strategy stuff
#include "pos_table.h"			// position table stuff
#include "seed_search.h"		// seed hit search stuff
#include "gapped_extend.h"		// gapped alignment stuff
#include "masking.h"			// dynamic masking stuff

// establish ownership of global variables

#ifdef lastz_owner
#define global
#else
#define global extern
#endif

//----------
//
// data structures and types
//
//----------

global int debug;				// how much debug info to show the programmer
								//   0   => nothing
								//   100 => everything


typedef struct infcontrol
	{
	char*	inferFileName;		// Name of file to write inferred scores to.  If
								// .. this includes "%s", then each score set we
								// .. try during convergence will be written,
								// .. with %s replaced by a string of the form
								// .. sNNN or gNNN.  The final score set will
								// .. be written with %s replaced by an empty
								// .. string
	score	inferScale;			// The desired value for the maximum subsitution
								// .. score.  If this is zero, no scaling is
								// .. performed.
	int		writeAsInt;			// true => we should produce scores that are
								//         .. integers even if the score type
								//         .. is floating-point
	int		hspThresholdIsRatio;// non-zero => the hspThreshold and
	int		gappedThresholdIsRatio;//  .. gappedThreshold value(s) passed into
								//     .. drive_scoring_inference is to be used
								//     .. as a ratio, as per ratioXXX value
	int		gapOpenIsRatio;		// non-zero => the scoring->gapOpen and
	int		gapExtendIsRatio;	//     .. scoring->gapExtend value(s) passed
								//     .. into drive_scoring_inference is to be
								//     .. used as a ratio, as per ratioXXX value
	int		subIterations;		// max number of iterations for inferring
								// .. substitution scores
	int		gapIterations;		// max number of iterations for inferring gap
								// .. scores
	int		idIsPercentile;		// true => control.minIdentity and
								//         .. control.maxIdentity are
								//         .. percentiles
								// false => minIdentity and maxIdentity are
								//         .. the actual percentages
	} infcontrol;

enum
	{
	ratioNone = 0,
	ratioMaxSubScore,
	ratioMinSubScore,
	};

// control--
//	data structure for providing global information to all parts of the lastz
//	program;  most of these are control parameters, but some information about
//	the problem being solved is also passed here
//
//	items marked here with an (I) are also valid for control of the scoring
//	inference process;  items marked (I-only) are *only* valid for inference;
//	items marked (I-copy) are valid for inference but are copies of pointers
//	to memory that belongs to the regular control set

typedef struct control
	{
	// the sequences being processed

	char*		seq1Filename;	// target sequence
	struct seq*	seq1;

	char*		seq2Filename;	// query sequence
	struct seq*	seq2;

	//	control parameters

	int			inferScores;	//     true => infer scores from the sequences
	int			inferOnly;		//     true => prohibit alignment (other than
								//             .. what is performed for
								//             .. inference)
	infcontrol	ic;				// (I-only) Additional inference controls.

	int			selfCompare;	// (I) true => seq1 and seq2 are the same
								//             .. sequence
	int			doSeedSearch;	// (I) false => prohibit the seed search (and
								//              .. prohibit alignment too)
	const s8*	charToBits;		//     table to map sequence characters to
								//     .. two-bit values, and illegal characters
								//     .. to -1;  indexed by a u8 value, 0..255
	int			whichStrand;	// (I) 0   => search + strand only
								//     < 0 => search - strand only
								//     > 0 => search both strands of target
	int			backward;		//     true => also search the reverse of
								//     .. strands;  reverse is just reverse,
								//     .. without complement;  this can be used
								//     .. to get an estimate of the
								//     .. false-positive rate of an alignment
	u32			step;			// (I) positional step size, indicating how
								//     .. often to store target word positions

	struct seed* hitSeed;		// (I) seeding strategy for hits
	int			maxIndexBits;	// (I) maximum number of index bits to use for
								//     .. the position table;  if the seed
								//     .. weight (in bits) is larger than this,
								//     .. the additional bits (effective hash
								//     .. collisions) are resolved using the
								//     .. sequences

	int			withTrans;		// (I) number of allowed transitions in a seed
								//     .. hit (0, 1 or 2)
	int			noHitFiltering;	//     true => just report every raw seed hit,
								//             .. with no filtering or
								//             .. processing
	u32			twinMinSpan;	// (I) span threshold for hits to be considered
	u32			twinMaxSpan;	// (I) .. twins;  twinMinSpan<=0 means we don't
								//     .. don't require twins (and we ignore
								//     .. twinMaxSpan);  see the twininfo struct
								//     .. in seed_search.h for details
	int			basicHitType;	// (I) the type of hit we require for a basic,
								//     .. single hit;  one of the hitXXX values
								//     .. below
	int			minMatches;		// (I) filter criteria for each seed hit;  we
	int			maxTransversions;//(I) .. require at least minMatches matches
								//     .. and no more than maxTransversions
								//     .. transversions;  if minMatches<0 no
								//     .. filtering is performed

	int			gfExtend;		//     whether to extend seed hits into HSPs
								//     .. (one of gfexXXX, from seed_search.h)
	int			mergeAnchors;	//     true => after seed hits/HSPs are found,
								//             .. merge any overlapping segments
	int			chain;			// (I) true => perform chaining
	score		chainDiag;		// (I) diagonal chaining penalty
	score		chainAnti;		// (I) antidiagonal chaining penalty
	int			gappedExtend;	// (I) true => perform gapped extension

	scoreset*	scoring;		// (I) scoring set
	scoreset*	maskedScoring;	// (I) scoring set with lowercase penalized;  in
								//     .. general, we treat lower case as bad
								//     .. during the search for HSPs, then treat
								//     .. upper/lower case as equivalent during
								//     .. the gapped alignment stage
	score		xDrop;			// (I) threshold to stop extensions of
								//     .. *ungapped* matches;  if the score drops
								//     .. off by more than xDrop, extension stops
	score		yDrop;			// (I) threshold to stop extensions of *gapped*
								//     .. alignments (similar to xdrop)
	sthresh		hspThreshold;	// (I) threshold for high scoring segment pairs
								//     .. (ungapped matches);  an HSP is
								//     .. discarded if its score is less than
								//     .. this threshold
	sthresh		gappedThreshold;// (I) threshold for high scoring alignments;  a
								//     .. gapped alignment is discarded if its
								//     .. score is less than this threshold
	int			entropicHsp;	// (I) true => involve entropy in the decision
								//             .. to discard an HSP
	int			reportEntropy;	//     true => report any HSPs that are
								//             .. discarded due to entropy
	int			mirror;			// (I) true => each HSP or seed hit is output
								//             .. with its symmetric copy,
								//             .. reflected over the diagonal
	u32			tracebackMem;	//     number of bytes to allocate to track
								//     .. gapped alignment traceback
	u8			dynamicMasking;	//     mask any position in target hit this many
								//     .. times;  zero or negative indicates no
								//     .. masking
	int			reportCensus;	//     true => report how many times each base
								//             .. in the target is aligned
	float		minIdentity;	// (I) the range of percent identity of HSPS we
	float		maxIdentity;	// (I) .. will keep (or alignment blocks of we
								//     .. are doing gapped extension);  these
								//     .. are values between 0.0 and 1.0
	float		minCoverage;	// (I) the range of query coverage of HSPS we
	float		maxCoverage;	// (I) .. will keep (or alignment blocks of we
								//     .. are doing gapped extension);  these
								//     .. are values between 0.0 and 1.0

	int			outputFormat;	//     format in which to write alignments;  one
								//     .. of fmtXXX values below
	int			outputNeedsTrueLen;//  true => the outputFormat requires that
								//             .. seq->trueLen is correct

	// for inner alignment (interpolation)

	score		innerThreshold;	//     threshold for HSPs during interpolation
								//     .. stage (equivalent to hspThreshold);
								//     .. zero or negative indicates no
								//     .. interpolation
	struct seed* innerSeed;		//     seeding strategy for inner alignment hits
	u32			innerWindow;	//     windowSize interpolation parameter (see
								//     .. inner_interpolate() and inner.c)

	// for quantum DNA query

	int			queryIsQuantum;	//     true => the query is quantum DNA
	score		ballScore;		//     minimum score required of a DNA word to
								//     .. be considered 'in' the ball
								//     .. surrounding a given quantum word;
								//     .. -1 indicates that this was never set
	qcode*		coding;			//     the code used for the query sequence

	// debug/info flags

	int			lajCompatible;	//     true => backward compatibility for LAJ
	u32			textContext;	//     the number of extra bp to print at the
								//     .. end of matches for outputFormat=fmtText
	char*		args;			//     a copy of the argv options, less sequence
								//     .. names, suitable for printing
	int			verbosity;		// (I) how much info to bombard the user with
								//       0   => minimal
								//       10  => everything
	int			reportTiming;	// (I) true => report runtime to the output file
								//             .. (e.g as z records in a gfa
								//             ..  file)
	int			reportStats;	// (I) true => report search statistics to the
								//             .. output file (e.g as z records
								//             .. in a gfa file)
	int			showStats;		// (I) true => show search statistics
	FILE*		statsFile;		// (I-copy) file to write statistics to (can be
								//     .. NULL)
	char*		statsFileName;	//     name of stats file (can be NULL)
	int			showPosTable;	//     whether or nto to show target positions
								//     .. table
								//       0 => don't show target positions table
								//       1 => show target positions table
								//       2 => show target positions table counts
								//            .. only
	} control;

// values for control parameters

enum
	{
	hitBad  = -1,
	hit_min = 0,
	hitSimple = hit_min,		// simple hit, no hash collision detection
	hitRecover,					// simple hit with hash collision recovery
	hit_max = hitRecover
	};


enum
	{
	fmtBad  = -1,
	fmt_min = 0,
	fmtGfa = fmt_min,			// output alignments in GFA format
	fmtGfaNoScore,				// .. in GFA format but without scores
	fmtLav,						// .. in LAV format
	fmtLavComment,				// .. in LAV format with comments
	fmtLavScore,				// .. in LAV format with scores where pctid is
	fmtLavText,					// .. in LAV format with as-text
	fmtAxt,						// .. in AXT format
	fmtAxtComment,				// .. in AXT format with comments
	fmtMaf,						// .. in MAF format
	fmtMafComment,				// .. in MAF format with comments
	fmtMafNoComment,			// .. in MAF format with no comments at all
	fmtText,					// .. as text
	fmtHspComp,					// .. as text, showing composition of each HSP
	fmtInfStats,				// .. as scoring inference stats
	fmtIdDist,					// .. as identity distribution
	fmtDeseed,					// .. as text for deseed program
	fmtInfScores,				// .. collect scoring inference stats (this
								// .. cannot be directly chosen by the user)
	fmtLavInfScores,			// .. fmtLav + fmtInfScores (debugging only)
	fmtNone,					// don't bother to output
	fmt_max = fmtNone
	};

#ifdef lastz_owner
char* formatNames[] = {"GFA","GFANOSCORE",
                       "LAV","lav+","LAVSCORE","lav+text",
                       "AXT","axt+","MAF","maf+","maf-",
                       "text",
                       "comp",
                       "infstats","iddist","deseed",
                       "infscores","lav+infscores",
                       "none" };
#else
extern char* formatNames[];
#endif

//----------
//
// prototypes for routines in lastz.c
//
//----------

void set_up_hit_processor (control* params, int collectingCensus,
                           hitprocessor* hitProc, void** hitProcInfo);

void start_one_strand     (seq* target, postable* targPositions, seq* query,
                           int emptyAnchors,
                           hitprocessor hitProc, void* hitProcInfo);
void finish_one_strand    (seq* target, u8* targetRev,
                           postable* targPositions,
                           seq* query,  u8* queryRev,
                           tback* traceback, census* targCensus);
void split_anchors        (int id);
void swap_anchors         (void);
void print_job_header     (void);

//----------
//
// statistics for events in this module
//
//----------

#if ((defined collect_stats) && (defined lastz_owner))

struct
	{
	float runTime;
	} lastzStats;

// stats macros

#define lastz_count_stat(field)   ++lastzStats.field
#define lastz_uncount_stat(field) --lastzStats.field
#define lastz_set_stat(field,val) (lastzStats.field = val)
#else
#define lastz_count_stat(field)
#define lastz_uncount_stat(field)
#define lastz_set_stat(field,val)
#endif // collect_stats && lastz_owner

#undef global
#endif // lastz_H
